import asyncio
from functools import lru_cache
import inspect
import threading
import uuid
import time
from typing import Optional, Callable, Any, Dict, Union, List

from PySide6.QtWidgets import QVBoxLayout, QWidget, QHBoxLayout, QPlainTextEdit, QLineEdit, QApplication, QStyledItemDelegate, \
        QSizePolicy, QPushButton, QSpinBox, QDoubleSpinBox, QGraphicsOpacityEffect, QFrame, QGroupBox
from PySide6.QtCore import Qt, QPoint, Signal, QDate, QTime, QTimer, QEvent, QSize
from PySide6.QtGui import QCursor

from qtmui.hooks import State
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style
from qtmui.utils.data import deep_merge
from qtmui.utils.translator import getTranslatedText
from ..li import Li

from ..menu import Menu
from ..menu_item import MenuItem

from .flow_layout import FlowLayout
from .tags_view import TagsView

from ..picker import DatePicker, TimePicker, AMTimePicker, ZhDatePicker, CalendarPicker, ColorPicker

from qtmui.hooks import useState

from qtmui.material.styles import useTheme
from qtmui.i18n.use_translation import translate, i18n
from ..utils._misc import signals_blocked
from ..utils.functions import _get_fn_args

from ...qtmui_assets import QTMUI_ASSETS

from ..py_iconify import PyIconify

from .py_combobox import PyComboBox
from .py_date_edit import PyDateEdit
from .py_date_time_edit import PyDateTimeEdit
from .py_dial import PyDial
from .py_double_spinbox import PyDoubleSpinBox
from .py_font_combobox import PyFontComboBox
from .tf_line_edit import TFLineEdit
from .tf_line_edit_multiple import TFLineEditMultiple
from .py_plaintext_edit import PyPlainTextEdit
from .py_plaintext_edit_multiple import PyPlainTextEditMultiple
from .py_spin_box import PySpinBox
from .py_text_edit import PyTextEdit
from .py_time_edit import PyTimeEdit


from ..widget_base import PyWidgetBase

from ..utils.validate_params import _validate_param

class TextField(QGroupBox):

    showPlaceholder = Signal()
    valueChanged = Signal(object)
    updatePopupPosition = Signal()
    createMenu = Signal()
    renderSearchMenuItems = Signal(object)
    
    updateStyleSheet = Signal(object)

    VALID_COLORS = ['primary', 'secondary', 'error', 'info', 'success', 'warning']
    VALID_MARGINS = ['dense', 'none', 'normal']
    VALID_SIZES = ['medium', 'small']
    VALID_VARIANTS = ['filled', 'outlined', 'standard']

    def __init__(
        self,
        key: str= None,
        autoComplete: Optional[Union[State, bool, str]] = None,
        autoFocus: Optional[Union[State, bool]] = False,
        autoWidth: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        classes: Optional[Union[State, Dict]] = None,
        color: Optional[Union[State, str]] = "primary",
        defaultOpen: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        defaultValue: Optional[Union[State, object]] = None,
        disabled: Optional[Union[State, bool]] = False,
        error: Optional[Union[State, bool]] = False,
        FormHelperTextProps: Optional[Union[State, Dict]] = None,
        fullWidth: Optional[Union[State, bool]] = False,
        helperText: Optional[Union[State, str]] = None,
        id: Optional[Union[State, str]] = None,
        InputLabelProps: Optional[Union[State, Dict]] = None,
        inputProps: Optional[Union[State, Dict]] = None,
        InputProps: Optional[Union[State, Dict]] = None,
        inputRef: Optional[Union[State, object]] = None,
        inputValue: Optional[Union[State, object]] = None,
        label: Optional[Union[State, str, Callable]] = None,
        max: Optional[Union[State, int, float]] = None,
        margin: Optional[Union[State, str]] = "none",
        maxRows: Optional[Union[State, int, str]] = None,
        min: Optional[Union[State, int, float]] = None,
        minRows: Optional[Union[State, int, str]] = None,
        multiline: Optional[Union[State, bool]] = False,
        multiple: Optional[Union[State, bool]] = False,
        menuHeight: Optional[Union[State, int]] = 400,
        name: Optional[Union[State, str]] = None,
        getOptionLabel: Optional[Union[State, Callable]] = None,
        open: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        onOpen: Optional[Union[State, Callable]] = None, # chỉ dùng cho Select
        onClose: Optional[Union[State, Callable]] = None, # chỉ dùng cho Select
        onChange: Optional[Union[State, Callable]] = None,
        onInputChange: Optional[Union[State, Callable]] = None,
        onMousePress: Optional[Union[State, Callable]] = None,
        options: Optional[Union[State, object]] = None,
        hiddenLabel: Optional[Union[State, bool]] = False,
        placeholder: Optional[Union[State, str]] = None,
        required: Optional[Union[State, bool]] = False,
        rows: Optional[Union[State, int, str]] = None,
        select: Optional[Union[State, bool]] = False,
        selected: Optional[Union[State, bool]] = False,
        selectOptions: Optional[Union[State, List]] = None,
        SelectProps: Optional[Union[State, Dict]] = None,
        size: Optional[Union[State, str]] = "medium",
        slotProps: Optional[Union[State, Dict]] = None,
        slots: Optional[Union[State, Dict]] = None,
        sx: Optional[Union[State, Dict, Callable, str]] = None,
        step: Optional[Union[State, int, float]] = None,
        renderSelectOptions: Optional[Union[State, Callable]] = None,
        renderOption: Optional[Union[State, Callable]] = None,
        renderValue: Optional[Union[State, Callable]] = None,
        renderTags: Optional[Union[State, Callable]] = None,
        readOnly: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        type: Optional[Union[State, str]] = "text",
        treeView: Optional[Union[State, bool]] = False,
        value = None,
        variant: Optional[Union[State, str]] = "outlined",
        asynRenderQss: Optional[Union[State, bool]] = False,
        *args, **kwargs
    ) -> None:
        # super().__init__(*args, **kwargs)
        super().__init__()
        
        self._kwargs = {
            **kwargs,
            "variant": variant,
            "size": size,
        }
        
        self._setKwargs(self._kwargs)
        
        self.setObjectName(str(uuid.uuid4()))
        start_time = time.time()

        self._key = key
        
        if self._key == "proxyTypekyyyyyyyyyyyyyyyyy":
            print('ProfileTextField__________________', self._key)

        self._key = key
        self._autoComplete = autoComplete
        self._autoFocus = autoFocus
        self._classes = classes
        self._color = color
        self._defaultValue = defaultValue
        self._disabled = disabled
        self._error = error
        self._FormHelperTextProps = FormHelperTextProps
        self._fullWidth = fullWidth
        self._helperText = helperText
        self._id = id
        self._InputLabelProps = InputLabelProps
        self._inputProps = inputProps
        self._InputProps = InputProps
        self._inputRef = inputRef
        self._inputValue = inputValue
        self._label = label
        self._max = max
        self._margin = margin
        self._maxRows = maxRows
        self._min = min
        self._minRows = minRows
        self._multiline = multiline
        self._multiple = multiple
        self._menuHeight = menuHeight
        self._name = name
        self._getOptionLabel = getOptionLabel
        self._onChange = onChange
        self._onInputChange = onInputChange
        self._onMousePress = onMousePress
        self._options = options
        self._hiddenLabel = hiddenLabel
        self._placeholder = placeholder
        self._required = required
        self._rows = rows
        self._select = select
        self._selected = selected
        self._selectOptions = selectOptions
        self._SelectProps = SelectProps
        self._size = size
        self._slotProps = slotProps
        self._slots = slots
        self._sx = sx
        self._step = step
        self._renderSelectOptions = renderSelectOptions
        self._renderOption = renderOption
        self._renderValue = renderValue
        self._renderTags = renderTags
        self._type = type
        self._treeView = treeView
        self._value = value
        self._variant = variant
        
        self._asynRenderQss = asynRenderQss

        if self._fullWidth:
            if not self._sx:
                self._sx = {}
            self._sx.update({"width": "100%"}) # lý do tại sao đặt ở đây thì fullWidth mới có tác dụng thì chưa rõ

        if self._sx:
            self._setSxDict(self._sx)

        self.__init_ui()
        
        end_time = time.time()
        
        # print("time_loading_textfield___________________________", (end_time - start_time)/0.00013971328735351562)


    def _connect_signals_props(self):
        if isinstance(self._value, State):
            self._value.valueChanged.connect(self._set_data)

    def __init_ui(self):
        self._init_done = False
        self._data = None

        self._connect_signals_props()

        self.theme = useTheme()
        
        i18n.langChanged.connect(self._set_translated_label)

        if self._fullWidth:
            if not self._sx:
                self._sx = {}
            self._sx.update({"width": "100%"})


        self._tagsWidget = None
        self._selected_options = {}
        self._inputFieldHasFocusIn = False


        self._inputField = None
        self._popup = None
        self._popupShowing = False
        self._btnSelect = None

        self._isDateTimeType = self._type == "date" or self._type == "time" or self._type == "date-time" or self._type == "color"

        self._menu = None
        self._menuItems, self._setMenuItems = useState([])
        
        self.renderSearchMenuItems.connect(self._render_search_menu_items)


        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(0,0,0,0)
        
        self._focused, self._setFocused = useState(False)
        self._hovered, self._setHovered = useState(False)
        
        self._focused.valueChanged.connect(self._set_focused_prop)

        if self._type == "int":
            self._inputField: PySpinBox = PySpinBox
            # self._inputField.setValue = self._set_spinbox_value
        elif self._type == "float":
            self._inputField: PyDoubleSpinBox = PyDoubleSpinBox
        elif self._type == "date-time":
            self._inputField: PyDateTimeEdit = PyDateTimeEdit
        elif self._type == "date":
            self._inputField: PyDateEdit = PyDateEdit
        elif self._type == "time":
            self._inputField: PyTimeEdit = PyTimeEdit
        else: # email, password, select
            if self._multiline:
                self._inputField: PyPlainTextEdit = PyPlainTextEdit
            elif self._multiple:
                if self._renderTags:
                    self._inputField: PyPlainTextEditMultiple = PyPlainTextEditMultiple
                else:
                    self._inputField: TFLineEdit = TFLineEdit
            else:
                self._inputField: TFLineEdit = TFLineEdit

        self._inputField = self._inputField(
            label=self._label,
            onMouseEnter=self.onMouseEnter,
            onMouseLeave=self.onMouseLeave,
            onFocusIn=self.onFocusIn,
            onFocusOut=self.onFocusOut,
            onMouseRelease=self.onMouseRelease,
            onChange=self.onChange,
            variant=self._variant,
            size=self._size,
        )


        if self._type == "date":
            self._inputField.dateChanged.connect(self._set_date_picker)
        elif self._type == "time":
            self._inputField.timeChanged.connect(self._set_time_picker)
        elif self._type == "color":
            pass
        elif  self._type == "text" or self._select: # có thể lỗi ở đây
            # chỗ này nhằm mục đích tạo ra tính năng search autocomplete
            self._inputField.textChanged.connect(self._on_text_changed)

        self.layout().addWidget(self._inputField)

        self._setup_input_props()
        self._setupy_input_type()

        if self._defaultValue is not None:
            self._set_data(self._defaultValue, setText=True)

        if self._value is not None:
            self._set_data( self._value, setText=True)

        if self._disabled:
            self.setEnabled(False)

        PyWidgetBase._installTooltipFilter(self)


        self.destroyed.connect(lambda obj: self._onDestroy())


        # ASYN INIT
        QTimer.singleShot(0, self._scheduleSetupSelectMode)
        
        # chỗ này nghiên cứu lại kỹ hơn, chưa kiểm định lỗi, ít khi menu dài tới hàm trăm phần tử, nên k nhất thiết load trước
        # QTimer.singleShot(500, self._scheduleCraeteMenu) 

        self.theme.state.valueChanged.connect(self._onThemeChanged)
        # QTimer.singleShot(0, self._scheduleSetStyleSheet)
        # self._setStyleSheet()
        if self._asynRenderQss:
            self.updateStyleSheet.connect(self._update_stylesheet)
        else:
            self._setStyleSheet()
            # QTimer.singleShot(0, self._setStyleSheet)
        
        if self._multiple:
            QTimer.singleShot(0, self._scheduleSetupDefaultTags)
            
        
        self._init_done = True

    def _onDestroy(self, obj=None):
        # Cancel task nếu đang chạy
        if hasattr(self, "_setupDefaultTags") and self._setupDefaultTags and not self._setupDefaultTags.done():
            self._setupDefaultTags.cancel()
        if hasattr(self, "_setupStyleSheet") and self._setupStyleSheet and not self._setupStyleSheet.done():
            self._setupStyleSheet.cancel()
        if hasattr(self, "_setupSelectMode") and self._setupSelectMode and not self._setupSelectMode.done():
            self._setupSelectMode.cancel()
        if hasattr(self, "_setupMenu") and self._setupMenu and not self._setupMenu.done():
            self._setupMenu.cancel()
            
    def _onThemeChanged(self):
        if not self.isVisible():
            return
        QTimer.singleShot(0, self._scheduleSetStyleSheet)
            
    def _scheduleSetupDefaultTags(self):
        self._setupDefaultTags = asyncio.ensure_future(self._lazySetupDefautTags())

    async def _lazySetupDefautTags(self):
        if self._defaultValue:
            for option in self._defaultValue:
                key = self._getOptionLabel(option)
                self._selected_options.update({key: option})
            # print('self._selected_options______________', self._selected_options)
            self.tags = self._renderTags(self._selected_options, self.getTagProps)
            # self.tags = self._renderTags(self._defaultValue, self.getTagProps)
            self._setFlowLayout(FlowLayout(self, children=self.tags + [self._inputFieldMultiple]))
            
    def _scheduleSetStyleSheet(self):
        self._setupStyleSheet = asyncio.ensure_future(self._lazy_setStyleSheet())

    async def _lazy_setStyleSheet(self):
        self._setStyleSheet()
        
    def _scheduleSetupSelectMode(self):
        self._setupSelectMode = asyncio.ensure_future(self._lazy_setupSelectMode())

    async def _lazy_setupSelectMode(self):
        self._setup_select_mode()

    def _scheduleCraeteMenu(self):
        self._setupMenu = asyncio.ensure_future(self._lazy_create_menu())

    async def _lazy_create_menu(self):
        self._create_menu()

    def _setup_select_mode(self):
        if self._select:
            # self.updatePopupPosition.connect(self._update_popup_position)

            # đặt ở đây để có _selectedKeys truyền cho TFLineEditMultiple => setup placeholder đúng
            if self._options:
                self._selectedKeys, self._setSelectedKeys = useState(self._options[0] if isinstance(self._options[0], str) else self._options[0].get("value") or "")

                if self._multiple:
                    if self._defaultValue:
                        selectedKeys = ["has_defaultValue"] # để thiết lập placeholder cho đúng
                        self._selectedKeys, self._setSelectedKeys = useState(selectedKeys)
                else:
                    self._selectedKeys, self._setSelectedKeys = useState(
                        self._options[0] if isinstance(self._options[0], str) 
                        else self._getOptionLabel(self._options[0]) if self._getOptionLabel else self._options[0].get("value") or ""
                    )
            elif self._renderOption: # cho Select - select thi k render tag ma render text value
                self._selectedKeys, self._setSelectedKeys = useState(self._data)

            elif self._selectOptions: 
                self._selectedKeys, self._setSelectedKeys = useState(self._data)

            self.action_frame = QWidget(self)
            self.action_frame.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
            hlo_action_frame = QHBoxLayout(self.action_frame)
            hlo_action_frame.setContentsMargins(0,0,0,0)
            self.action_frame.setFixedWidth(54)
            self.action_frame.layout().setAlignment(Qt.AlignRight | Qt.AlignVCenter)
            # self.action_frame.setStyleSheet("border: 1px solid red;")
            # self.setup_action_frame_and_adornment()

            if self._multiple:
                if self._renderTags:
                    self._inputFieldMultiple = TFLineEditMultiple(
                        textField=self,
                        label=self._label,
                        placeholder=self._placeholder, 
                        size=self._size, 
                        selectedKeys=self._selectedKeys,
                        onMouseEnter=self.onMouseEnter,
                        onMouseLeave=self.onMouseLeave,
                        onFocusIn=self.onFocusInMultiple,
                        onFocusOut=self.onFocusOut,
                        onMouseRelease=self.onMouseRelease,
                        onChange=self.onChange,
                    )
                    self._inputFieldMultiple.textChanged.connect(self._on_text_changed)
                    self._inputFieldMultiple.installEventFilter(self)
                    self._flowLayout, self._setFlowLayout = useState(FlowLayout(self, [self._inputFieldMultiple]))
                    self._flow_layout_view = TagsView(content=self._flowLayout)
                    self._flow_layout_view.layout().setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
                    self._inputField.setReadOnly(True)
                    self._inputField.setLayout(QHBoxLayout())
                    self._inputField.layout().setContentsMargins(0,0,54,0)
                    self._inputField.setPlaceholderText("")
                    self._inputField.layout().addWidget(self._flow_layout_view)

                elif self._renderValue:
                    self._inputField.setText(self._renderValue(self._data))

                if isinstance(self._renderOption, State):
                    self._renderOption.valueChanged.connect(self._create_menu)
            else:
                if not self._isDateTimeType and not isinstance(self._inputField, QSpinBox)  and not isinstance(self._inputField, QDoubleSpinBox):
                    pass

            if self._options and not self._getOptionLabel:
                raise AttributeError("Opp!! 'getOptionLabel' is required attribute!")
            
            if self._InputProps and self._InputProps.get("type") == "search":
                self._btn_clear_text= QPushButton()
                self._btn_clear_text.setFixedSize(24, 24)
                self._btn_clear_text.setIcon(PyIconify(key=QTMUI_ASSETS.ICONS.CLOSE))
                self._btn_clear_text.setCursor(Qt.PointingHandCursor)
                self._btn_clear_text.clicked.connect(self._on_clear_text)
                self._btn_clear_text.hide()
                hlo_action_frame.addWidget(self._btn_clear_text)
                self._btn_clear_text.hide()
            else:
                self._btn_clear_text= QPushButton()
                self._btn_clear_text.setFixedSize(24, 24)
                self._btn_clear_text.setIcon(PyIconify(key=QTMUI_ASSETS.ICONS.CLOSE))
                self._btn_clear_text.setCursor(Qt.PointingHandCursor)
                self._btn_clear_text.clicked.connect(self._on_clear_text)
                self._btn_clear_text.hide()
                hlo_action_frame.addWidget(self._btn_clear_text)
                self._btn_clear_text.hide()
            self._btn_clear_text.setFocusPolicy(Qt.NoFocus)
            self._btn_clear_text.setAttribute(Qt.WA_ShowWithoutActivating, True)

            self._btnSelect = QPushButton()
            self._btnSelect.setCursor(Qt.PointingHandCursor)
            self._btnSelect.setFixedSize(24, 24)
            if self._isDateTimeType:
                self._btnSelect.hide()
                self._btnSelect.setIconSize(QSize(24, 24))
                
            self._btnSelect.setIcon(PyIconify(key=QTMUI_ASSETS.ICONS.ARROW_DOWN))
            self._btnSelect.clicked.connect(self._btn_select_toggle_popup)
            hlo_action_frame.addWidget(self._btnSelect)

            self._inputField.installEventFilter(self)

            if hasattr(self, '_flow_layout_view'):
                self._flow_layout_view.mousePressEvent = self.onMultipleInputFieldClicked
                
            self.setup_action_frame_and_adornment() # cần thiết khi không async setup stylesheet
            
    @classmethod
    def _setSxDict(cls, sx: dict = {}):
        cls.sxDict = sx
        
    @classmethod
    def _setKwargs(cls, kwargs: dict = {}):
        cls.ownerState = kwargs

    @classmethod
    @lru_cache(maxsize=128)
    def _getSxQss(cls, sxStr: str = "", className: str = "TextField"):
        sx_qss = get_qss_style(cls.sxDict, class_name=className)
        return sx_qss
    

    # @timer
    def _setStyleSheet(self):
        theme = useTheme()
        
        if self._multiple:
            MuiAutocomplete_styles = theme.components["MuiAutocomplete"].get("styles")
            MuiAutocomplete_styles_root_multiple_min_height = MuiAutocomplete_styles.get("root")["@multiple"]["min-height"]
            MuiAutocomplete_styles_root_multiple_with_chip_min_height = MuiAutocomplete_styles.get("root")["@multiple"]["@chip"]["min-height"]
            if self._renderTags:
                self.setMinimumHeight(MuiAutocomplete_styles_root_multiple_with_chip_min_height)
            else:
                self.setMinimumHeight(MuiAutocomplete_styles_root_multiple_min_height)
        
        self.setProperty("variant", self._variant)
        
        if not self._data:
            self.setProperty("hasValue", False)
        else:
            self.setProperty("hasValue", True)

        
        if self._renderTags:
            _renderTags = "True"
        else:
            _renderTags = "False"

        if hasattr(self, "styledDict"):
            styledConfig = str(self.styledDict)
        else:
            styledConfig = "TextField"
            
        stylesheet = self._getStylesheet(styledConfig, self._variant, self._size, theme.palette.mode, str(self._multiple), str(self._multiline), _renderTags, str(self._focused.value))
        
        sxQss = ""
        if self._sx:
            sxQss = self._getSxQss(sxStr=str(self._sx), className=f"#{self.objectName()}")
            # sxQss = self._getSxQss(sxStr=str(self._sx), className=f"TextField")
        
        stylesheet = f"""
            {stylesheet}
            {sxQss}
        """

        self.setStyleSheet(stylesheet)


    @classmethod
    @lru_cache(maxsize=128)
    def _getStylesheet(cls, styledConfig: str, _variant: str, _size: str, _theme_mode: str, _multiple: str, _multiline: str, _renderTags:str, _focused: str):
        """Set the stylesheet for the Box."""
        
        theme = useTheme()
        if hasattr(cls, "styledDict"):
            themeComponent = deep_merge(theme.components, cls.styledDict)
        else:
            themeComponent = theme.components
            
        # print("getttttttttttttttttttttttttttttttttttt..................")

        # Chung cho toàn bộ input
        MuiInput = themeComponent["MuiInput"].get("styles")
        MuiInput_root = MuiInput.get("root")
        MuiInput_root_qss = get_qss_style(MuiInput_root)
        MuiInput_title = MuiInput.get("title")
        MuiInput_title_qss = get_qss_style(MuiInput_title)
        MuiInput_title_slot_focused_qss = get_qss_style(MuiInput_title["slots"]["focused"])
        MuiInput_title_slot_error_qss = get_qss_style(MuiInput_title["slots"]["error"])
        MuiInput_title_slot_disabled_qss = get_qss_style(MuiInput_title["slots"]["disabled"])
        
        MuiInput_inputField = MuiInput.get("inputField")
        MuiInput_inputField_qss = get_qss_style(MuiInput_inputField)
        MuiInput_inputField_prop_multiline_qss = get_qss_style(MuiInput_inputField["props"]["multiline"][f"{_size}"])
        MuiInput_inputField_prop_multiple_qss = get_qss_style(MuiInput_inputField["props"]["multiple"])
        MuiInput_inputField_prop_filledVariant_size_qss = ""
        if _variant == "filled":
            MuiInput_inputField_prop_filledVariant_size_qss = get_qss_style(MuiInput_inputField["props"]["filledVariant"][_size])

        MuiInputSize = themeComponent["MuiInputSize"]["styles"][f"{_size}"]
        MuiInputSize_qss = get_qss_style(MuiInputSize)

        MuiInputSize_not_multiple_qss = get_qss_style(MuiInputSize["no-multiple"])
        MuiInputSize_no_multiple_height = MuiInputSize["no-multiple"]["height"]

        # thiết lập tuỳ theo variant
        MuiVariantInput = themeComponent[f"Mui{_variant.capitalize()}Input"].get("styles")
        MuiVariantInput_root = MuiVariantInput.get("root")
        MuiVariantInput_root_qss = get_qss_style(MuiVariantInput_root)
        MuiVariantInput_root_slot_hovered_qss = get_qss_style(MuiVariantInput_root["slots"]["hover"])
        MuiVariantInput_root_slot_focused_qss = get_qss_style(MuiVariantInput_root["slots"]["focus"])
        
        MuiVariantInput_inputField_root = MuiVariantInput.get("inputField")["root"]
        MuiVariantInput_inputField_root_qss = get_qss_style(MuiVariantInput_inputField_root)

        MuiVariantInput_root_prop_hasStartAdornment_qss = ""
        if MuiVariantInput_inputField_root.get("props"):
            MuiVariantInput_root_prop_hasStartAdornment_qss = get_qss_style(MuiVariantInput_inputField_root["props"]["hasStartAdornment"])

        MuiInputSize = themeComponent["MuiInputSize"]["styles"][f"{_size}"]
        MuiInputSize_qss = get_qss_style(MuiInputSize)
        
        MuiFilledInput_title_qss = ""
        if _variant == "filled":
            MuiFilledInput_title_qss = get_qss_style(MuiVariantInput["title"])

        # print('MuiInputSize_qss______________', MuiInputSize_qss)

        _________object_name_______ = "_________object_name_______"


        input_base_stylesheet = f"""
                TextField {{
                    {MuiInput_root_qss}
                    {MuiInputSize_qss}
                    {MuiInputSize_not_multiple_qss if not (_multiple=="True" and _renderTags=="True") and not _multiline=="True" else ""}
                }}

                TextField QComboBox, QDateEdit, QDateTimeEdit, QDial, QDoubleSpinBox, QFontComboBox,
                QLineEdit, QPlainTextEdit, QSpinBox, QTextEdit, QTimeEdit {{
                    {MuiInput_inputField_qss}
                }}
                TextField[multipleHasValue=true] QPlainTextEdit {{
                    {MuiInput_inputField_prop_multiple_qss}
                }}
                TextField QPlainTextEdit, QTextEdit{{
                    {MuiInput_inputField_prop_multiline_qss}
                }}
                TextField QPlainTextEdit{{
                    min-height: 40px;
                }}
                TextField::title {{
                    {MuiInput_title_qss}
                    {MuiFilledInput_title_qss}
                }}
                TextField[focused=true]::title {{
                    {MuiInput_title_slot_focused_qss}
                }}
                TextField[error=true]::title {{
                    {MuiInput_title_slot_error_qss}
                }}
                TextField[disabled=true]::title {{
                    {MuiInput_title_slot_disabled_qss}
                }}
            """

        _variant_stylesheet = f"""
                TextField[variant={_variant}] {{
                    {MuiVariantInput_root_qss}
                    {MuiInputSize_qss}
                }}
                TextField[variant={_variant}]  QLineEdit, QPlainTextEdit, QTextEdit {{
                    {MuiVariantInput_inputField_root_qss}
                }}
                TextField[outlinedHasStartAdornment=true]  QLineEdit, QPlainTextEdit, QTextEdit {{
                    {MuiVariantInput_root_prop_hasStartAdornment_qss}
                }}
                TextField[hovered=true] {{
                    {MuiVariantInput_root_slot_hovered_qss}
                }}
                TextField[focused=true] {{
                    {MuiVariantInput_root_slot_focused_qss}
                }}
                TextField[focused=true] QLineEdit, QPlainTextEdit, QTextEdit {{
                    {MuiInput_inputField_prop_filledVariant_size_qss}
                }}
                TextField[hasValue=true] QLineEdit, QPlainTextEdit, QTextEdit {{
                    {MuiInput_inputField_prop_filledVariant_size_qss}
                }}
            """
        
        return input_base_stylesheet + _variant_stylesheet

    def _render_stylesheet(self):
        _theme_mode = useTheme().palette.mode
        self.setProperty("variant", self._variant)
        if self._renderTags:
            _renderTags = "True"
        else:
            _renderTags = "False"

        stylesheet = self._getStylesheet(self._variant, self._size, _theme_mode, str(self._multiple), str(self._multiline), _renderTags )
        
        sx_qss = ""
        if self._sx:
            if isinstance(self._sx, dict):
                sx_qss = get_qss_style(self._sx, class_name=f"#{self.objectName()}")
            elif isinstance(self._sx, Callable):
                sx = self._sx()
                if isinstance(sx, dict):
                    sx_qss = get_qss_style(sx, class_name=f"#{self.objectName()}")
                elif isinstance(sx, str):
                    sx_qss = sx
            elif isinstance(self._sx, str) and self._sx != "":
                sx_qss = self._sx
                
        self.updateStyleSheet.emit(stylesheet + sx_qss)

    def _update_stylesheet(self, stylesheet):
        self.setStyleSheet(stylesheet)

    def _create_menu(self):
        print("_create_menu called")

        if not self._menu:
            self._menu = Menu(
                maxHeight=self._menuHeight,
                children=self._menuItems
            )

        if isinstance(self._renderOption, State): # dành cho Select, option được tạo luôn khi gọi renderOption
            if isinstance(self._renderOption.value, Callable):
                self._selectOptions = self._renderOption.value()
                for option in self._selectOptions:
                    if isinstance(option, Li):
                        option._set_selected_keys(self._selectedKeys)
                        option._set_on_changed(self._selected_change)

        elif self._selectOptions and len(self._selectOptions) and isinstance(self._selectOptions[0], QWidget):
            for option in self._selectOptions:
                if isinstance(option, MenuItem):
                    if hasattr(self, '_selectedKeys'):
                        option._set_selected_keys(self._selectedKeys)
                    option._set_on_changed(self._selected_change)

        elif self._options:
            self._selectOptions = []
            for _option in self._options:
                props = {}
                option = self._renderOption(props, _option)

                if isinstance(option, Li):
                    option._set_selected_keys(self._selectedKeys)
                    option._set_on_changed(self._selected_change)
                self._selectOptions.append(option)

        elif self._renderOption and isinstance(self._renderOption, Callable):
            self._selectOptions = self._renderOption(**(_get_fn_args(self._renderOption)))
            for option in self._selectOptions:
                if isinstance(option, Li):
                    option._set_selected_keys(self._selectedKeys)
                    option._set_on_changed(self._selected_change)

        self._setMenuItems(self._selectOptions)
      
            
    def onFocusIn(self, event=None):
        self._setFocused(True)
        self._show_label()

        if hasattr(self, '_inputFieldMultiple'):
                self._inputFieldMultiple.setFocus(Qt.OtherFocusReason)
            
    def onFocusInMultiple(self, event=None):
        self.onFocusIn(event)

    def onFocusOut(self, event=None):
        print('onFocusOut______________5555555_______')
        
        # dành cho thiết lập khi nhấn vào nút clear hay select
        if hasattr(self, '_btn_clear_text') and QApplication.widgetAt(QCursor.pos()) is self._btn_clear_text:
            try:
                QTimer.singleShot(0, lambda: self._inputField.setFocus(Qt.OtherFocusReason))
            except Exception as e:
                print(f"Error setting focus: {e}")
            return

        if hasattr(self, '_btnSelect') and QApplication.widgetAt(QCursor.pos()) is self._btnSelect:
            try:
                QTimer.singleShot(0, lambda: self._inputField.setFocus(Qt.OtherFocusReason))
            except Exception as e:
                print(f"Error setting focus: {e}")
            return
        
        if hasattr(self, '_inputFieldMultiple') and QApplication.widgetAt(QCursor.pos()) is self._inputFieldMultiple:
            try:
                QTimer.singleShot(0, lambda: self._inputFieldMultiple.setFocus(Qt.OtherFocusReason))
            except Exception as e:
                print(f"Error setting focus: {e}")
            return
        
        self._setFocused(False)
        print('onFocusOut______________77777777777_______')
        

        # nếu inputValue hiện tại không nằm trong options thì clear nó đi
        if self._select:
            if self._multiple:
                if len(self._selectedKeys.value) == 0:
                    self._inputFieldMultiple.setText("")
                    self._inputFieldMultiple._set_placer_holder_text()
                    self._hide_label()
            else:
                if isinstance(self._inputField, TFLineEdit):
                    current_input_value = self._inputField.text()
                elif isinstance(self._inputField, PyPlainTextEdit):
                    current_input_value = self._inputField.toPlainText()
                    
                option_values = []
                options = self._options
                if options:
                    for option in options:
                        label = option
                        if self._getOptionLabel:
                            label = self._getOptionLabel(option)
                        option_values.append(label.lower())
                if current_input_value.lower() not in option_values:
                    self._set_data(None)
                    if isinstance(self._inputField, TFLineEdit):
                        self._inputField.setText("")
                    elif isinstance(self._inputField, PyPlainTextEdit):
                        self._inputField.setPlainText("")
                    self._inputField._set_placer_holder_text()
                else:
                    # set lại data dựa trên inputValue
                    for option in options:
                        label = option
                        if self._getOptionLabel:
                            label = self._getOptionLabel(option)
                        if label.lower() == current_input_value.lower():
                            self._set_data(option)
                            self._inputField.setText(label)
                            break

        # if self._data is None:
        if not self._data:
            self._hide_label()

        if not self.underMouse():
            if hasattr(self, "_btn_clear_text") and self._btn_clear_text.isVisible():
                self._btn_clear_text.setVisible(False)


            

    def onMouseEnter(self, event=None):
        self._setHovered(True)
        if self._data:
            if hasattr(self, "_btn_clear_text") and not self._btn_clear_text.isVisible():
                self._btn_clear_text.setVisible(True)
        
    def onMouseLeave(self, event=None):
        self._setHovered(False)

    def onMouseRelease(self, event=None):
        if self._onMousePress:
            self._onMousePress()
        if self._select:
            if not self._isDateTimeType and not isinstance(self._inputField, QPlainTextEdit):
                if self._inputField.selectedText() != "":
                    return

    def onMultipleInputFieldClicked(self, event=None):
        if self._onMousePress:
            self._onMousePress()
        if self._select:
            if not self._isDateTimeType and not isinstance(self._inputField, QPlainTextEdit):
                if self._inputField.selectedText() != "":
                    return
            self._toggle_popup()

    def onChange(self, value=None):
        # nếu là select thì chưa set_value
        if not self._select:
            self._set_data(value, setText=False)

        if self._onInputChange:
            self._onInputChange(value)

    def _set_state(self, stateValue: dict):
        value = stateValue.get("value")
        state = stateValue.get("state")

    def _set_data(self, value=None, valueChanged=True, setText=True):
        
        _selected_options = self._selected_options.copy()
        
        for key, val in _selected_options.items():
            if isinstance(value, list) and val not in value:
                del self._selected_options[key]
                
        if isinstance(value, State):
            self._data = value.value
        else:
            self._data = value

        if self._type == "int":
            pass
        if self._type == "int": # PySpinBox
            pass
        elif self._type == "float":
            pass
        elif self._type == "date-time":
            pass
        elif self._type == "date":
            pass
        elif self._type == "time":
            pass
        elif self._multiline:
            pass
        elif self._multiple:
            if self._data is not None:
                self._set_multiple_has_value_prop(True)
            else:
                self._set_multiple_has_value_prop(False)

        else: # email, password, select
            pass
            # if self._multiline or self._multiple:
            #     self._inputField = PyPlainTextEdit
            # else: # TFLineEdit
            #     if self._data is not None:
            #         if self._variant == "filled":
            #             self._set_filled_has_value_prop(True) # làm cho inputField padding top 1 khoảng cho giống mui JS
            #     else:
            #         if self._variant == "filled":
            #             self._set_filled_has_value_prop(False) # làm cho inputField padding top 1 khoảng cho giống mui JS

        if value is not None: # value == "" cũng vào đây
            if self._data != "":
                self._show_label()

        if valueChanged:
            self.valueChanged.emit(self._data)

        if setText:
            self._set_text()
            


    def _set_text(self): # value
        if self._data == None:
            if self._type == "text":
                if isinstance(self._inputField, QPlainTextEdit):
                    self._inputField.setPlainText("")
                elif isinstance(self._inputField, QLineEdit):
                    self._inputField.setText("")
            return
        
        if self._data == "" and not self._select:
            if isinstance(self._inputField, QPlainTextEdit):
                self._inputField.setPlainText("")
            elif isinstance(self._inputField, QLineEdit):
                self._inputField.setText("")
            return

        if self._select:
            if not self._multiple: # multiple thì set tag không cần set text
                if isinstance(self._data, dict):
                    if self._getOptionLabel:
                        text = self._getOptionLabel(self._data)
                        self._inputField.setText(str(text))
                else:
                    if self._getOptionLabel:
                        for option in self._options:
                            if isinstance(option, dict):
                                if self._getOptionLabel(option) == self._data: # options khong phai dạng value label
                                    text = self._getOptionLabel(option)
                                    self._inputField.setText(str(text))
                                    break
                            elif isinstance(option, str):
                                if option == self._data:
                                    text = self._getOptionLabel(option)
                                    self._inputField.setText(str(text))
                                    break
                    else:
                        self._inputField.setText(str(self._data))

            else: # multiple
                if self._renderValue:
                    self._inputField.setText(self._renderValue(self._data))

        else:
            if self._multiline:
                if isinstance(self._data, dict):
                    if self._getOptionLabel:
                        text = self._getOptionLabel(self._data)
                        self._inputField.setPlainText(text)
                else:
                    self._inputField.setPlainText(str(self._data))
            else:
                if isinstance(self._data, dict):
                    if self._getOptionLabel:
                        text = self._getOptionLabel(self._data)
                        self._inputField.setText(text)
                else:
                    if self._type == "date":
                        if isinstance(self._data, str):
                            with signals_blocked(self._inputField):
                                self._inputField.setDate(QDate.fromString(self._data, self._inputField.displayFormat()))
                        else:
                            self._inputField.setDate(self._data)
                    elif self._type == "time":
                        if isinstance(self._data, str):
                            self._inputField.setTime(QTime.fromString(self._data, self._inputField.displayFormat()))
                        else:
                            self._inputField.setTime(self._data)
                    elif self._type == "int":
                        self._inputField.setValue(int(self._data)) 
                    elif self._type == "float":
                        self._inputField.setValue(float(self._data)) 
                    else:
                        if isinstance(self._inputField, QPlainTextEdit):
                            self._inputField.setPlainText(self._data) 
                        if isinstance(self._inputField, QLineEdit):
                            self._inputField.setText(str(self._data)) 


    def _set_slot(self, data):
        if data.get("slot") == "error":
            self._inputField.setProperty("p-error", "true")
        elif data.get("slot") == "valid":
            self._inputField.setProperty("p-error", "false")

    def _on_text_changed(self, s=None):
        # print("s: ", s)
        
        if not self._init_done:
            return

        # Callback khi input thay đổi. Lấy dữ liệu người dùng gõ tuỳ ý mà chưa cần chọn ngay để kiểm soát hoàn toàn quá trình lọc, reset, hoặc validation.
        if self._onInputChange:
            self._onInputChange(s)

        if not self._autoComplete:
            return
        
        if s is None:
            return
        
        if s == "":
            self._selected_options = {}

        if self._InputProps and self._InputProps.get("type") == "search":
            # thay doi icon clear text
            pass
        
        if self._multiline:
            if self._inputField.toPlainText() == "":
                self._btn_clear_text.hide()
            else:
                self._btn_clear_text.show()
        elif self._multiple:
            if self._inputFieldMultiple.text() == "":
                self._btn_clear_text.hide()
            else:
                self._btn_clear_text.show() 
        else:
            if self._inputField.text() == "":
                self._btn_clear_text.hide()
            else:
                self._btn_clear_text.show() 

        # tìm kiếm các key phù hợp kết quả tìm kiếm
        currentValue = None
        if isinstance(self._data, dict):
            if self._getOptionLabel:
                currentValue = self._getOptionLabel(self._data)
        else:
            currentValue = self._data

        # if s != currentValue and self._InputProps and self._InputProps.get("type") == "search":
        if s != currentValue or s == "" and self._init_done:
            search_options_results = []
            if self._options:
                for option in self._options:
                    if isinstance(option, dict):
                        if self._getOptionLabel:
                            key = self._getOptionLabel(option)
                            if s in key:
                                search_options_results.append(option)
                    else:
                        key = option
                        if s in key:
                            search_options_results.append(option)

            self.renderSearchMenuItems.emit(search_options_results)

        if self._error:
            self._error = None
            
        if self._init_done:
            self._show_popup()

    def _render_search_menu_items(self, search_options_results):
        search_menu_items = []
        if len(search_options_results):
            for _option in search_options_results:
                # props = {'text': self._getOptionLabel(_option)}
                props = {}
                option = self._renderOption(props, _option)
                if isinstance(option, Li):
                    option._set_selected_keys(self._selectedKeys)
                    option._set_on_changed(self._selected_change)
                search_menu_items.append(option)
            if not len(search_menu_items):
                search_menu_items = [Li(key="No options", selected=True, value=">>>>>>No options")]
        else:
            search_menu_items = [Li(text="No options", selected=True, value=">>>>>>No options")]
            
        self._setMenuItems(search_menu_items)

    def validate_type_int(self):
        # Kiểm tra tính hợp lệ của min và max
        if self._min is not None and self._max is not None:
            if not isinstance(self._min, int) or not isinstance(self._max, int):
                raise ValueError("Both 'min' and 'max' must be integers.")
            if self._max < self._min:
                raise ValueError("'max' must be greater than or equal to 'min'.")
        else:
            raise ValueError("Both 'min' and 'max' required.")
        return True

    def _setupy_input_type(self):
        if self._type == "int":
            if self._min:
                if not isinstance(self._min, int):
                    raise ValueError("'min' must be integer.")
                self._inputField.setMinimum(self._min)
            if self._max:
                if not isinstance(self._max, int):
                    raise ValueError("'max' must be integer.")
                self._inputField.setMaximum(self._max)
            if self._min and self._max:
                if self._max < self._min:
                    raise ValueError("'max' must be greater than or equal to 'min'.")
            if self._step:
                if not isinstance(self._step, int):
                    raise ValueError("'step' must be integer.")
                self._inputField.setSingleStep(self._step)
        elif self._type == "float":
            if self._min:
                if not isinstance(self._min, float):
                    raise ValueError("'min' must be float.")
                self._inputField.setMinimum(self._min)
            if self._max:
                if not isinstance(self._max, float):
                    raise ValueError("'max' must be float.")
                self._inputField.setMaximum(self._max)
            if self._min and self._max:
                if self._max < self._min:
                    raise ValueError("'max' must be greater than or equal to 'min'.")
            if self._step:
                if not isinstance(self._step, int) or not isinstance(self._step, float):
                    raise ValueError("'step' must be number (int or float).")
                self._inputField.setSingleStep(self._step)

        elif self._type == "password":
            self._inputField.setEchoMode(QLineEdit.Password) 
        elif self._type == "email":
            pass

    def _set_date_picker(self, date: QDate):
        if isinstance(self.startAdornment, DatePicker) or isinstance(self.startAdornment, ZhDatePicker):
            self.startAdornment.setDate(date)
            self._data = date.toString()
            self.valueChanged.emit(self._data)

    def _set_time_picker(self, time: QTime):
        if isinstance(self.startAdornment, TimePicker) or isinstance(self.startAdornment, AMTimePicker):
            self.startAdornment.setTime(time)
            self._data = time.toString()
            self.valueChanged.emit(self._data)

    def _setup_input_props(self):
        if self._InputProps:
            if self._InputProps.get('startAdornment') and isinstance(self._InputProps.get('startAdornment'), QWidget):
                self.startAdornment = self._InputProps.get('startAdornment')
                self.layout().insertWidget(0, self.startAdornment)
                self._set_prop_has_start_adornment(True)
                self._inputField.setProperty('has-start-adornment', "true")
                if isinstance(self.startAdornment, DatePicker) or isinstance(self.startAdornment, ZhDatePicker):
                    self.startAdornment.dateChanged.connect(self._set_data)
                    if isinstance(self.startAdornment, ZhDatePicker):
                        self._inputField.setDisplayFormat("yyyy/MM/dd")
                    else:
                        self._inputField.setDisplayFormat("dd/MM/yyyy")
                    
                if isinstance(self.startAdornment, AMTimePicker) or isinstance(self.startAdornment, TimePicker):
                    self.startAdornment.timeChanged.connect(self._set_data)
                    if isinstance(self.startAdornment, AMTimePicker):
                        self._inputField.setDisplayFormat(f"hh:mm{'{}'.format(':ss') if self.startAdornment.isSecondVisible() else ''} ap")
                    else:
                        self._inputField.setDisplayFormat(f"hh:mm{'{}'.format(':ss') if self.startAdornment.isSecondVisible() else ''}")

            if self._InputProps.get('endAdornment') and isinstance(self._InputProps.get('endAdornment'), QWidget):
                self.endAdornment = self._InputProps.get('endAdornment')
                self.layout().insertWidget(-1, self.endAdornment)

    def _on_clear_text(self):
        self._onClearText = True

        if self._multiple:
            self._setSelectedKeys([])
            self._btn_clear_text.hide()
            self._selected_options = {}
            self._set_data([])
            self._setFlowLayout(FlowLayout(self, children=[self._inputFieldMultiple]))
        elif self._multiline:
            self._set_data("")
            self._inputField.setPlainText("")
        elif self._type == "date":
            self._set_data(None)
            self._inputField.setDate(QDate.currentDate())
        elif self._type == "time":
            self._set_data(None)
            self._inputField.setTime(QTime.currentTime())
        elif self._type == "int":
            self._set_data(None)
            self._inputField.setValue(0)
        elif self._type == "float":
            self._set_data(None)
            self._inputField.setValue(0.0)
        elif self._select:
            self._set_data(None)
            self._inputField.setText("")
        else: # text, email, password
            self._set_data("")
            self._inputField.setText("")
            
        self._btn_clear_text.hide()
        QTimer.singleShot(0, lambda: self._inputField.setFocus(Qt.OtherFocusReason))
        
        if self._onChange:
            self._onChange(self._data)
            
        if self._onInputChange:
            self._onInputChange("")
        

    def _update_popup_position(self):
        if hasattr(self, "_popup") and self._popup is not None:
            popup_position = self.mapToGlobal(QPoint(0, self.height()))
            self._popup.setGeometry(popup_position.x(), popup_position.y(), self.width(), 50)


    def _hide_popup(self):
        if self._select:
            if self._popup and self._popup.isVisible():
                self._popup.setVisible(False)
            if self._isDateTimeType:
                icon = PyIconify(key=QTMUI_ASSETS.ICONS.CALENDAR)
            else:
                icon = PyIconify(key=QTMUI_ASSETS.ICONS.ARROW_DOWN)

            if self._btnSelect:
                self._btnSelect.setIcon(icon)
        self._popupShowing = False


    def _btn_select_toggle_popup(self, e=None):
        if self._popup and self._popupShowing:
            print('__btn_select_toggle_popup_______________hide_popup')
            self._hide_popup()
        else:
            print('__btn_select_toggle_popup______________show_popup')
            self._show_popup()

    def _toggle_popup(self, e=None):
        if self._select and not self._popupShowing:
            self._show_popup()
            return
        if self._select and (self._data or self._data == "") and self._inputField.hasFocus():
            return
        if self._popup and self._popupShowing:
            self._hide_popup()
        else:
            self._show_popup()

    def _show_popup(self, search=None):
        
        if not self._menu:
            self._create_menu()
        
        if self._popup and self._popupShowing:
            return

        if not self._popup:
            self._popup = QFrame(self)
            self._popup.setFocusPolicy(Qt.NoFocus)
            self._popup.setAttribute(Qt.WA_ShowWithoutActivating, True)
            self._popup.setWindowFlags(Qt.Window |  Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
            self._popup.setAttribute(Qt.WA_TranslucentBackground, True) # Enable translucent background => setStyleSheet ko có tác dụng gì
            QApplication.instance().installEventFilter(self)
            self._popup.setObjectName(str(uuid.uuid4()))
            self._popup.setLayout(QVBoxLayout())
            self._popup.layout().setContentsMargins(0,0,9,9)
            self._popup.layout().addWidget(self._menu)
            self._popup.setAutoFillBackground(False)
            # self._popup.setStyleSheet(f"#{self._popup.objectName()} {{background: transparent;border: 2px solid red;}}")

        # Lấy điểm cục bộ (x, y) của relativeTo và chuyển sang global
        global_point = self.mapToGlobal(QPoint(0, self.height()))
        
        # tính toán vị trí hiển thị popup
        main_window_pos_y = QApplication.instance().mainWindow.height()
        self_pos_y = global_point.y()
        
        self._popup.move(-3000, -3000) # di chuyển tạm để lấy kích thước
        
        self._popup.setVisible(True)
        # External WM_DESTROY received for  QWidgetWindow(0x1e1758d8fd0, name="QFrameClassWindow") , parent:  QWindow(0x0) , transient parent:  QWindow(0x0)
        # QWindowsBackingStore::flush: GetDC failed (Invalid window handle.)
        # QTimer.singleShot(100, self._popup.show)
        

        if (main_window_pos_y - self_pos_y) < self._menu.height():
            global_point.setY(global_point.y() - self.height() - self._menu.height())
        
        self._popup.move(global_point)
        self._popup.setMinimumWidth(self.width())

        self._menu.setVisible(True)

        if self._isDateTimeType:
            icon = PyIconify(key=QTMUI_ASSETS.ICONS.CALENDAR)
        else:
            icon = PyIconify(key=QTMUI_ASSETS.ICONS.ARROW_UP)
        self._btnSelect.setIcon(icon)
        self._popupShowing = True
        

    def _on_menu_item_selected(self, key):
        self._hide_popup()
        self._data = key

    def _about_to_hide(self):
        self._popupShowing = False


    def getTagProps(self, index):
        return {"onDelete": lambda key=index: self._on_tag_deleted(key)}

    def _on_tag_deleted(self, index):
        self._selected_change(index)

    def _update_selected_options(self, key=None, option=None):
        """
        Hàm này nhằm cập nhật giá trị self._selected_options cho Select
        """
        if key not in list(self._selected_options.values()):
            self._selected_options.update({key: key})
        else:
            self._selected_options.pop(key)

    def _setup_tags(self, key=None, option=None):
        if key:
            for _option in self._options:
                if self._getOptionLabel(_option) == key:
                    option = _option
                    break
        else:
            if option:
                if isinstance(option, dict):
                    key = self._getOptionLabel(option)

        if option not in self._selected_options.values():
            try:
                self._selected_options.update({key: option})
                self.tags = self._renderTags(self._selected_options, self.getTagProps)
            except Exception as e:
                import traceback
                traceback.print_exc()
            self._setFlowLayout(FlowLayout(self, children=self.tags + [self._inputFieldMultiple]))
        else:
            self._selected_options.pop(key)
            self.tags = self._renderTags(self._selected_options, self.getTagProps)
            self._setFlowLayout(FlowLayout(self, children=self.tags + [self._inputFieldMultiple]))

    def _selected_change(self, data):

        if isinstance(data, dict):
            key = data.get("value")
            value = data.get("value")
        else:
            key = data
            value = data

        if self._multiple:
            # for delete option item
            if key in self._selectedKeys.value:
                selected_keys = self._selectedKeys.value.copy()
                selected_keys.remove(key)
                self._setSelectedKeys(selected_keys)
                if self._renderTags:
                    self._setup_tags(key=key)
                elif self._renderValue:
                    self._update_selected_options(key)
                values = []
                for selected_option in self._selected_options.values():
                    values.append(selected_option)

                self._set_data(values)

                if self._onChange:
                    self._onChange(self._data)

                if self._renderValue:
                    self._inputField.setText(self._renderValue(self._data))

                return
            
            if self._renderTags: # cho Autocomplete - render tag
                self._setup_tags(key=key, option=key)
            else: # cho Select
                self._update_selected_options(key)
            
            keys = []
            values = []
            for selected_option in self._selected_options.values():
                if isinstance(selected_option, dict):
                    values.append(selected_option)
                    if self._getOptionLabel:
                        key = self._getOptionLabel(selected_option)
                        keys.append(key)
                elif isinstance(selected_option, str):
                    values.append(selected_option)
                    keys.append(selected_option)

            self._setSelectedKeys(keys)
            self._set_data(values)

            if self._onChange:
                self._onChange(self._data)

            if self._renderValue:
                self._inputField.setText(self._renderValue(self._data))

            self._hide_popup()
        else: # not multiple
            self._setSelectedKeys(key)
            self._set_data(value)
            self._selected_options.update({key: key})                
            self._hide_popup()

            # Call onChange with the new value
            if self._onChange:
                # self._onChange(data) # contronled state mode không đúng, nhận vào list mà onchange lại là dict
                self._onChange(self._data)
                
        if self._menu:
            for item in self._menu.findChildren(Li):
                item.show()

    def _show_label(self):
        if not self._hiddenLabel:
            self._set_translated_label()

    def _set_translated_label(self):
        if not self._hiddenLabel:
            self.setTitle(getTranslatedText(self._label))

    def _hide_label(self):
        print('_hide_label')
        self.setTitle("")

    def _set_hovered_prop(self):
        if self._disabled:
            return
        if isinstance(self._hovered, State):
            self.setProperty("hovered", self._hovered.value)
        elif isinstance(self._hovered, bool):
            self.setProperty("hovered", self._hovered)
        
    def _set_focused_prop(self):
        if isinstance(self._focused, State):
            self.setProperty("focused", self._focused.value)
        elif isinstance(self._focused, bool):
            self.setProperty("focused", self._focused)
        self._setStyleSheet()
        
    def _set_multiple_has_value_prop(self, state):
        self.setProperty("multipleHasValue", state)
        
    def _set_prop_has_start_adornment(self, state:bool):
        self.setProperty("hasStartAdornment", state)

    def setup_action_frame_and_adornment(self):
        if self._InputProps and self._InputProps.get('startAdornment') and isinstance(self._InputProps.get('startAdornment'), QWidget):
            pass
        if self._InputProps and self._InputProps.get('endAdornment') and isinstance(self._InputProps.get('endAdornment'), QWidget):
            pass
        
        if hasattr(self, 'action_frame'):
            self.action_frame.move(self.width() - self.action_frame.width(), (self.height() + 6 - self.action_frame.height()) // 2)
            if not self.action_frame.isVisible():
                self.action_frame.show()

    def leaveEvent(self, event):
        if not self.underMouse() and not self._inputField.hasFocus():
            if hasattr(self, "_btn_clear_text") and self._btn_clear_text.isVisible():
                self._btn_clear_text.setVisible(False)
        return super().leaveEvent(event)

    def resizeEvent(self, event):
        self.setup_action_frame_and_adornment()
        return super().resizeEvent(event)
    
    def showEvent(self, event):
        if self._asynRenderQss:
            threading.Thread(target=self._render_stylesheet, args=(), daemon=True).start()
        return super().showEvent(event)
    
    def eventFilter(self, obj, event):
        if self._select:
            if obj == self._inputField or hasattr(self, '_inputFieldMultiple') and obj == self._inputFieldMultiple:
                if event.type() == QEvent.MouseButtonPress:
                    self._show_popup()
            elif self._popup and self._popup.isVisible():
                if event.type() == QEvent.MouseButtonPress and not self.underMouse():
                    if not self._popup.geometry().contains(event.globalPos()):
                        self._hide_popup()
                elif event.type() == QEvent.Wheel and ((not self.underMouse() and not self._popup.underMouse()) or obj == self._inputField):
                    self._hide_popup()
            
        return super().eventFilter(obj, event)




import asyncio
from functools import lru_cache
import uuid
from typing import Optional, Callable, Any, Dict, Union

from PySide6.QtWidgets import QSpinBox, QHBoxLayout
from PySide6.QtCore import QTimer
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style
from  .py_input_base import MuiInputBase

from qtmui.material.styles import useTheme
from qtmui.hooks import State
from qtmui.i18n.use_translation import translate

class PySpinBox(QSpinBox, MuiInputBase):
    def __init__(
        self,
        parent=None,
        label: Optional[str] = None,
        onChange: Optional[Callable] = None,
        children: Optional[list] = None,  # Hỗ trợ định nghĩa ghi đè hệ thống cũng như các kiểu CSS bổ sung.
        sx: Optional[Union[Dict, Callable, str]] = None,  # Hỗ trợ định nghĩa ghi đè hệ thống cũng như các kiểu CSS bổ sung.
        variant: str = "outlined",
        size: str = "medium",  # Kích thước của thành phần ('medium' hoặc 'small').
        **kwargs
    ):
        super().__init__()
        self.setObjectName(str(uuid.uuid4()))
        # self._setUpUi(**kwargs)

        self._children = children
        self._label = label
        self._onChange = onChange
        self._sx = sx
        self._variant = variant
        self._size = size
        
        self._init_ui()
        
    def _init_ui(self):
        self.theme = useTheme()

        if isinstance(self._children, list):
            self.setLayout(QHBoxLayout())
            self.layout().setContentsMargins(0,0,0,0)
            for widget in self._children:
                self.layout().addWidget(widget)

        # self.textChanged.connect(self._onChange)

        self.theme.state.valueChanged.connect(self._onThemeChanged)
        # QTimer.singleShot(0, self._scheduleSetStyleSheet)
        self._set_stylesheet()
        
        self.destroyed.connect(lambda obj: self._onDestroy())
        

    def _onDestroy(self, obj=None):
        # Cancel task nếu đang chạy
        if hasattr(self, "_setupStyleSheet") and self._setupStyleSheet and not self._setupStyleSheet.done():
            self._setupStyleSheet.cancel()

    def _onThemeChanged(self):
        if not self.isVisible():
            return
        QTimer.singleShot(0, self._scheduleSetStyleSheet)

    def _scheduleSetStyleSheet(self):
        self._setupStyleSheet = asyncio.ensure_future(self._lazy_set_stylesheet())

    async def _lazy_set_stylesheet(self):
        self._set_stylesheet()


    @classmethod
    @lru_cache(maxsize=128)
    def _get_stylesheet(cls, _variant: str, _size: str, _theme_mode: str):
        
        theme = useTheme()
        PySpinBox_root = theme.components["PySpinBox"].get("styles").get("root")
        PySpinBox_root_qss = get_qss_style(PySpinBox_root)

        _________object_name_______ = "_________object_name_______"

        stylesheet = f"""
                #{_________object_name_______}[variant={_variant}] {{
                    {PySpinBox_root_qss}
                }}
            """
        return stylesheet
    
    def _set_stylesheet(self, component_styled=None):
        _theme_mode = useTheme().palette.mode
        self.setProperty("variant", self._variant)
        stylesheet = self._get_stylesheet(self._variant, self._size, _theme_mode)
        stylesheet = stylesheet.replace("_________object_name_______", self.objectName())
        sx_qss = ""

        if self._sx:
            if isinstance(self._sx, dict):
                sx_qss = get_qss_style(self._sx, class_name=f"#{self.objectName()}")
            elif isinstance(self._sx, Callable):
                sx = self._sx()
                if isinstance(sx, dict):
                    sx_qss = get_qss_style(sx, class_name=f"#{self.objectName()}")
                elif isinstance(sx, str):
                    sx_qss = sx
            elif isinstance(self._sx, str) and self._sx != "":
                sx_qss = self._sx

        # print('stylesheet___________', stylesheet)
        self.setStyleSheet(stylesheet + sx_qss)
        
        
    def enterEvent(self, event):
        MuiInputBase.enterEvent(self, event)
        super().enterEvent(event)
    
    def leaveEvent(self, event):
        MuiInputBase.leaveEvent(self, event)
        super().leaveEvent(event)
    
    def mousePressEvent(self, event):
        MuiInputBase.mousePressEvent(self, event)
        super().mousePressEvent(event)
    
    def mouseReleaseEvent(self, event):
        MuiInputBase.mouseReleaseEvent(self, event)
        super().mouseReleaseEvent(event)
    
    def focusInEvent(self, event) -> None:
        MuiInputBase.focusInEvent(self, event)
        super().focusInEvent(event)

    def focusOutEvent(self, event) -> None:
        MuiInputBase.focusOutEvent(self, event)
        super().focusOutEvent(event)
    


from typing import TYPE_CHECKING, Optional

from ....system.color_manipulator import alpha

from .properties_name import *


if TYPE_CHECKING:
    from ..theme_reducer import ThemeState


def alert(_theme):
    theme: ThemeState = _theme

    COLORS = ['info', 'success', 'warning', 'error']

    lightMode = theme.palette.mode == 'light'

    return {
        'PyAlertStandard': {
            'styles': {
                'root': {
                    f"{_color}": {
                        border: "0px solid transparent",
                        borderRadius: "8px",
                        color: getattr(theme.palette, _color).darker if lightMode else getattr(theme.palette, _color).lighter,
                        backgroundColor: getattr(theme.palette, _color).lighter if lightMode else getattr(theme.palette, _color).darker,
                    } 
                    for _color in COLORS
                },
                'icon': {
                    f"{_color}": {
                        color: getattr(theme.palette, _color).main if lightMode else getattr(theme.palette, _color).lighter,
                        opacity: 1,
                    } 
                    for _color in COLORS
                },

            },
        },
        'PyAlertFilled': {
            'styles': {
                'root': {
                    f"{_color}": {
                        border: "0px solid transparent",
                        borderRadius: "8px",
                        color: getattr(theme.palette, _color).contrastText,
                        backgroundColor: getattr(theme.palette, _color).main,
                    } 
                    for _color in COLORS
                },
                'icon': {
                    opacity: 1,
                },

            },
        },
        'PyAlertOutlined': {
            'styles': {
                'root': {
                    f"{_color}": {
                        backgroundColor: alpha(getattr(theme.palette, _color).main, 0.08),
                        color: getattr(theme.palette, _color).dark if lightMode else getattr(theme.palette, _color).light,
                        border: f"solid 1px {alpha(getattr(theme.palette, _color).main, 0.16)}",
                        borderRadius: "8px",
                    } 
                    for _color in COLORS
                },
                'icon': {
                    f"{_color}": {
                        color: getattr(theme.palette, _color).main,
                        opacity: 1,
                    } 
                    for _color in COLORS
                },
            },
        },
        'PyAlertTitle': {
            'styles': {
                'root': {
                    f"{_color}": {
                        color: getattr(theme.palette, _color).main,
                        marginBottom: theme.spacing(0.5),
                        fontWeight: 700,
                    } 
                    for _color in COLORS
                },
            },
        },
    }



from .qtrangeslider import QRangeSlider
from PySide6 import QtCore
from PySide6 import QtWidgets as QtW

QSS = """
                QSlider:horizontal {
                        min-height: 24px;
                }

                QSlider::groove:horizontal {
                        height: 4px;
                        background-color: rgba(0, 0, 0, 100);
                        border-radius: 2px;
                }

                QSlider::sub-page:horizontal {
                        background: #00A76F;
                        height: 4px;
                        border-radius: 2px;
                }

                QSlider::handle:horizontal {
                        border: 1px solid rgb(222, 222, 222);
                        width: 20px;
                        min-height: 24px;
                        margin: -9px 0;
                        border-radius: 11px;
                        background-color: qradialgradient(spread:pad, cx:0.5, cy:0.5, radius:0.5, fx:0.5, fy:0.5,
                                stop:0 #00A76F,
                                stop:0.48 #00A76F,
                                stop:0.55 rgb(255, 255, 255),
                                stop:1 rgb(255, 255, 255));
                }

                QSlider::handle:horizontal:hover {
                        background-color: qradialgradient(spread:pad, cx:0.5, cy:0.5, radius:0.5, fx:0.5, fy:0.5,
                                stop:0 #00A76F,
                                stop:0.55 #00A76F,
                                stop:0.65 rgb(255, 255, 255),
                                stop:1 rgb(255, 255, 255));
                }

                QSlider::handle:horizontal:pressed {
                        background-color: qradialgradient(spread:pad, cx:0.5, cy:0.5, radius:0.5, fx:0.5, fy:0.5,
                                stop:0 #00A76F,
                                stop:0.4 #00A76F,
                                stop:0.5 rgb(255, 255, 255),
                                stop:1 rgb(255, 255, 255));
                }

                QSlider::groove:horizontal:disabled {
                        background-color: rgba(0, 0, 0, 75);
                }

                QSlider::handle:horizontal:disabled {
                        background-color: #808080;
                        border: 5px solid #cccccc;
                }


                QSlider:vertical {
                        min-width: 24px;
                }

                QSlider::groove:vertical {
                        width: 4px;
                        background-color: rgba(0, 0, 0, 100);
                        border-radius: 2px;
                }

                QSlider::add-page:vertical {
                        background: #00A76F;
                        width: 4px;
                        border-radius: 2px;
                }

                QSlider::handle:vertical {
                        border: 1px solid rgb(222, 222, 222);
                        height: 20px;
                        min-width: 24px;
                        margin: 0 -9px;
                        border-radius: 11px;
                        background-color: qradialgradient(spread:pad, cx:0.5, cy:0.5, radius:0.5, fx:0.5, fy:0.5,
                                stop:0 #00A76F,
                                stop:0.48 #00A76F,
                                stop:0.55 rgb(255, 255, 255),
                                stop:1 rgb(255, 255, 255));
                }

                QSlider::handle:vertical:hover {
                        background-color: qradialgradient(spread:pad, cx:0.5, cy:0.5, radius:0.5, fx:0.5, fy:0.5,
                                stop:0 #00A76F,
                                stop:0.55 #00A76F,
                                stop:0.65 rgb(255, 255, 255),
                                stop:1 rgb(255, 255, 255));
                }

                QSlider::handle:vertical:pressed {
                        background-color: qradialgradient(spread:pad, cx:0.5, cy:0.5, radius:0.5, fx:0.5, fy:0.5,
                                stop:0 #00A76F,
                                stop:0.4 #00A76F,
                                stop:0.5 rgb(255, 255, 255),
                                stop:1 rgb(255, 255, 255));
                }

                QSlider::groove:vertical:disabled {
                        background-color: rgba(0, 0, 0, 75);
                }

                QSlider::handle:vertical:disabled {
                        background-color: #808080;
                        border: 5px solid #cccccc;
                }
        """

SYSTEM_QSS = """
                QSlider::handle:horizontal {
                        background-color: #00A76F;
                }
                QSlider::handle:vertical {
                        background-color: #00A76F;
                }

        """


class DemoWidget(QtW.QWidget):
    def __init__(self) -> None:
        super().__init__()

        reg_hslider = QtW.QSlider(QtCore.Qt.Horizontal)
        reg_hslider.setValue(50)
        range_hslider = QRangeSlider(QtCore.Qt.Horizontal)
        range_hslider.setValue((20, 80))
        multi_range_hslider = QRangeSlider(QtCore.Qt.Horizontal)
        multi_range_hslider.setValue((11, 33, 66, 88))
        multi_range_hslider.setTickPosition(QtW.QSlider.TicksAbove)
        reg_hslider.setStyleSheet(SYSTEM_QSS)
        range_hslider.setStyleSheet(SYSTEM_QSS)
        multi_range_hslider.setStyleSheet(SYSTEM_QSS)

        styled_reg_hslider = QtW.QSlider(QtCore.Qt.Horizontal)
        styled_reg_hslider.setValue(50)
        styled_reg_hslider.setStyleSheet(QSS)
        styled_range_hslider = QRangeSlider(QtCore.Qt.Horizontal)
        styled_range_hslider.setValue((20, 80))
        styled_range_hslider.setStyleSheet(QSS)

        reg_vslider = QtW.QSlider(QtCore.Qt.Vertical)
        reg_vslider.setValue(50)
        range_vslider = QRangeSlider(QtCore.Qt.Vertical)
        range_vslider.setValue((22, 77))
        reg_vslider.setStyleSheet(SYSTEM_QSS)
        range_vslider.setStyleSheet(SYSTEM_QSS)

        tick_vslider = QtW.QSlider(QtCore.Qt.Vertical)
        tick_vslider.setValue(55)
        tick_vslider.setTickPosition(QtW.QSlider.TicksRight)
        tick_vslider.setStyleSheet(SYSTEM_QSS)

        range_tick_vslider = QRangeSlider(QtCore.Qt.Vertical)
        range_tick_vslider.setValue((22, 77))
        range_tick_vslider.setTickPosition(QtW.QSlider.TicksLeft)
        range_tick_vslider.setStyleSheet(SYSTEM_QSS)

        szp = QtW.QSizePolicy.Maximum
        left = QtW.QWidget()
        left.setLayout(QtW.QVBoxLayout())
        left.setContentsMargins(2, 2, 2, 2)
        label1 = QtW.QLabel("Regular QSlider Unstyled")
        label2 = QtW.QLabel("QRangeSliders Unstyled")
        label3 = QtW.QLabel("Styled Sliders (using same stylesheet)")
        label1.setSizePolicy(szp, szp)
        label2.setSizePolicy(szp, szp)
        label3.setSizePolicy(szp, szp)
        left.layout().addWidget(label1)
        left.layout().addWidget(reg_hslider)
        left.layout().addWidget(label2)
        left.layout().addWidget(range_hslider)
        left.layout().addWidget(multi_range_hslider)
        left.layout().addWidget(label3)
        # left.layout().addWidget(styled_reg_hslider)
        # left.layout().addWidget(styled_range_hslider)

        right = QtW.QWidget()
        right.setLayout(QtW.QHBoxLayout())
        right.setContentsMargins(15, 5, 5, 0)
        right.layout().setSpacing(30)
        right.layout().addWidget(reg_vslider)
        right.layout().addWidget(range_vslider)
        right.layout().addWidget(tick_vslider)
        right.layout().addWidget(range_tick_vslider)

        self.setLayout(QtW.QHBoxLayout())
        self.layout().addWidget(left)
        self.layout().addWidget(right)
        self.setGeometry(600, 300, 580, 300)
        self.activateWindow()
        # self.show()




import uuid
from typing import Callable, Optional, Sequence, Union

from PySide6.QtCore import Property, Signal, Slot
from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel

from qtmui.hooks import useState

from qtmui.material.styles import useTheme
from qtmui.hooks import State

from ..controller import Controller
from ..textfield import TextField
from ..form_control import FormControl
from ..select import Select
from ..box import Box
from ..chip import Chip


from ..qss_name import *

class RHFSelect(QWidget):
    valueChanged = Signal(object)

    def __init__(
            self,
            name: str,
            label: Optional[Union[str, State, Callable]]=None,
            native: bool = None,
            maxHeight: int = 220,
            onChange: Optional[Callable] = None,
            PaperPropsSx: object = None,
            helperText: str = None,
            children: object = None,
            selected: bool = False,
            **kwargs
            ):
        super().__init__()

        self._name = name
        self._label = label
        self._native = native
        self._maxHeight = maxHeight
        self._onChange = onChange
        self._children = children
        self._PaperPropsSx = PaperPropsSx
        self._helperText = helperText
        self._selected = selected
        self._value = None

        self._state, self._setState = useState(None)


        self.setLayout(QVBoxLayout())
        self.setObjectName(str(uuid.uuid4()))
        # self.setStyleSheet("background: green;") # str multi line
        control = False
        error = {}

        self._stateSignal = None

        self._select = Select(
            name=self._name,
            # selected=self._selected,
            onChange=lambda newValue:
                self.set_value(newValue)
            ,
            label=self._label,
            children=children,
            **kwargs
        )

        self.layout().addWidget(
            Controller(
                name=name,
                control=control,
                render=self._select
            )
        )

        self.lbl_helper_text = QLabel(self)
        self.lbl_helper_text.setStyleSheet(f'''
            padding-left: 8px;
            {useTheme().typography.caption.to_qss_props()}
            font-size: 11px;
        ''')
        self.lbl_helper_text.setText("")
        self.layout().addWidget(self.lbl_helper_text)

        # self._complete.valueChanged.connect(self.set_value)

        # if self._onValueChanged is not None:
        #     self.valueChanged.connect(self._onValueChanged)


    def set_value(self, value=None):
        self._value = value
        if self._value == []:
            self._value = None
        # print(f"RHFAutocomplete {self._name}: {self._value}")
        if isinstance(self._selected, State):
            self._selected.value = True

        if self._onChange:
            self._onChange(value)
        self.valueChanged.emit(value)

        self._select._inputField._set_value(value=self._value, valueChanged=False)
        
        self._setState(self._value)

    @Property(bool)
    def stateSignal(self):
        return self._stateSignal

    @stateSignal.setter
    def stateSignal(self, value):
        self._stateSignal = value
        self._stateSignal.connect(self.state)

    @Slot(object)
    def state(self, state):
        theme = useTheme()
        if self._name == state.get("field"):
            if state.get("error"):
                self._select._inputField._set_slot({"slot": "error", "message": state.get("error_message")})
                self.lbl_helper_text.setText(str(state.get("error_message")[0]))
                self.lbl_helper_text.setStyleSheet(f'''
                    padding-left: 8px;
                    font-size: {theme.typography.caption.fontSize};
                    font-weight: {theme.typography.caption.fontWeight};
                    color: {theme.palette.error.main};
                ''')
                if not self.lbl_helper_text.isVisible():
                    self.lbl_helper_text.show()
            else:
                self._select._inputField._set_slot({"slot": "valid"})
                self.lbl_helper_text.hide()


class RHFMultiSelect(QWidget):
    def __init__(
            self,
            name: str = None,
            chip: object = None,
            label: str = None,
            options: object = None,
            checkbox: object = None,
            placeholder: str = None,
            helperText: str = None,
            sx: dict = None
            ):
        super().__init__()
        self.setLayout(QVBoxLayout())
        # self.layout().setContentsMargins(0,0,0,0)
        self.setObjectName(str(uuid.uuid4()))
        # self.setStyleSheet('''#{}  {{ {} }}'''.format(self.objectName(), "background: green;")) # str multi line
        self.setStyleSheet("background: green;") # str multi line

        control = False
        error = {
            "message": "ksjdkl"
        }
        field = {
            "value": "option 1"
        }

        self.layout().addWidget(
            Controller(
                name=name,
                control=control,
                render=FormControl(
                    sx=sx,
                    children=[
                        # InputLabel(label=label) if label else None,
                        # Select(
                        #     # **field,
                        #     multiple=True,
                        #     displayEmpty=not placeholder,
                        #     labelId=name,
                        #     input=OutlinedInput(fullWidth=True, select=True, label=label, error=not error),
                        #     renderValue=self.render_values,
                        #     children=[
                        #         MenuItem(enabled=False, value="", children=[QLabel('nnnnnnnnnnnnnnnnnnnnnnnnnnnn')]), # Typography(text=placeholder, variant="h6")
                        #         *[
                        #             MenuItem(
                        #                 text=option["label"],
                        #                 key=option["value"], 
                        #                 value=option["value"], 
                        #                 children=[
                        #                     Checkbox(size="small", disableRipple=True, checked=option["value"] in field["value"]) 
                        #                 ]
                        #             ) for option in options
                        #         ]
                        #     ]
                        # ),
                        # FormHelperText(error=error.get("message") if error and hasattr(error, 'message') else helperText)
                    ]
                ),
            )
        )

    def render_values(self, selected_ids, options, placeholder=None, chip=False):
        selected_items = [item for item in options if item['value'] in selected_ids]

        if not selected_items and placeholder:
            return Box(component='em', children=[QLabel(placeholder)])

        if chip:
            chips = []
            for item in selected_items:
                chips.append(Chip(item['label']))
            return Box(children=chips)

        return ', '.join(item['label'] for item in selected_items)



from PySide6.QtWidgets import QApplication, QWidget
from PySide6.QtCore import QTimer, QEasingCurve
from PySide6.QtGui import QPainter, QColor
import sys
import time

from qtmui.material.styles import useTheme

COLORS = ['default', 'primary', 'secondary', 'info', 'success', 'warning', 'error']

def alpha(color: QColor, alpha: float) -> QColor:
    """Trả về QColor với alpha (0..1)."""
    c = QColor(color)
    c.setAlphaF(alpha)  # hoặc setAlpha(int(alpha*255))
    return c

class LinearIndeterminate(QWidget):
    def __init__(
        self,
        key: str = None,
        color: str = "default",
    ):
        super().__init__()
        self.setMinimumHeight(5)
        self.setMaximumHeight(5)
        self.setMinimumWidth(300)
        
        self.theme = useTheme()
        
        if color == "default":
            self.value_color = self.theme.palette.text.secondary
        elif color in COLORS:
            self.value_color = getattr(self.theme.palette, color).main
        
        self.bg_color = alpha(self.value_color, 0.16)

        self.start_time = time.time()
        self.duration = 3.0  # thời gian 1 giai đoạn
        self.timer = QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(16)

        # tạo 2 easing curves cho 2 giai đoạn
        self.curve1 = QEasingCurve(QEasingCurve.InOutCubic)
        self.curve2 = QEasingCurve(QEasingCurve.InOutCubic)

    def paintEvent(self, event):
        w = self.width()
        h = self.height()
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.fillRect(0, 0, w, h, self.bg_color)

        elapsed = (time.time() - self.start_time)
        t = elapsed % (self.duration * 2)  # chu kỳ 2 giai đoạn

        if t < self.duration:
            # giai đoạn 1: width tăng từ 50% -> 66%, tốc độ InOut
            p = self.curve1.valueForProgress(t / self.duration)
            width_ratio = 0.5 + 0.16 * p  # width từ 50% -> 66%
            value_width = width_ratio * w
            # lùi sang trái bằng width hiện tại để thanh đi từ ngoài màn hình vào
            x = (w + value_width) * p - value_width
        else:
            # giai đoạn 2: width giảm từ 66% -> 33%
            p = self.curve2.valueForProgress((t - self.duration) / self.duration)
            width_ratio = 0.66 - 0.33 * p
            value_width = width_ratio * w
            x = (w + value_width) * p - value_width

        painter.fillRect(x, 0, value_width, h, self.value_color)

# if __name__ == "__main__":
#     app = QApplication(sys.argv)
#     w = LinearIndeterminate()
#     w.setWindowTitle("Linear Indeterminate Progress")
#     w.show()
#     sys.exit(app.exec())



# coding:utf-8
from typing import Union

from PySide6.QtCore import Property,QRectF,Qt,QPoint
from PySide6.QtGui import QIcon, QPainter,QColor
from PySide6.QtWidgets import QWidget
from PySide6.QtSvg import QSvgRenderer

from ...common.icon import FluentIconBase, drawIcon, toQIcon
from ...common.overload import singledispatchmethod


class IconWidget(QWidget):
    """ Icon widget

    Constructors
    ------------
    * IconWidget(`parent`: QWidget = None)
    * IconWidget(`icon`: QIcon | str | FluentIconBase, `parent`: QWidget = None)
    """

    @singledispatchmethod
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setIcon(QIcon())

    @__init__.register
    def _(self, icon: FluentIconBase, parent: QWidget = None):
        self.__init__(parent)
        self.setIcon(icon)

    @__init__.register
    def _(self, icon: QIcon, parent: QWidget = None):
        self.__init__(parent)
        self.setIcon(icon)

    @__init__.register
    def _(self, icon: str, parent: QWidget = None):
        self.__init__(parent)
        self.setIcon(icon)

    def getIcon(self):
        return toQIcon(self._icon)

    def setIcon(self, icon: Union[str, QIcon, FluentIconBase]):
        self._icon = icon
        self.update()

    def paintEvent(self, e):
        painter = QPainter(self)
        painter.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)
        drawIcon(self._icon, painter, self.rect())
        # if isinstance(self._icon,FluentIconBase):
        #     drawIcon(self._icon, painter, self.rect())
        # else:
        #     # painter.setBrush(QColor(255, 255, 255))
        #     # painter.setPen(Qt.NoPen)
        #     # painter.drawEllipse(QPoint(35 // 2, 35 // 2), 35 // 2, 35 // 2)
        #     renderer = QSvgRenderer(self._icon)
        #     renderer.render(painter, QRectF(self.rect()))
        

    icon = Property(QIcon, getIcon, setIcon)


from typing import Optional, Union, Dict, List, Callable
import uuid

from PySide6.QtWidgets import (
    QFrame, QWidget, QGraphicsDropShadowEffect, QVBoxLayout, QLabel
)
from PySide6.QtGui import QColor
from PySide6.QtCore import Qt, Property

from qtmui.material.system.color_manipulator import rgbaToQColor
from qtmui.material.typography import Typography
from qtmui.hooks import State
from qtmui.material.styles import useTheme
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style


class Paper(QFrame):
    def __init__(
        self,
        parent=None,
        children: Optional[Union[State, str, QWidget, List[QWidget]]] = None,
        classes=None,
        component: Optional[Union[State, QWidget]] = None,
        elevation: Union[State, int] = 1,
        square: Union[State, bool] = False,
        sx: Optional[Union[State, List, Dict, Callable, str]] = None,
        variant: Union[State, str] = "elevation",
        color=None,
        *args, **kwargs
    ):
        super().__init__(parent)
        self.setObjectName(str(uuid.uuid4()))

        # === Props ===
        self._classes = classes or {}
        self._component = component
        self._elevation = elevation
        self._square = square
        self._sx = sx
        self._variant = variant
        # self._color = QColor(color) if color else None
        self._color = color

        # === Layout ===
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        if children:
            if isinstance(children, list):
                for child in children:
                    if isinstance(child, QWidget):
                        self._layout.addWidget(child)
            elif isinstance(children, str):
                # self._layout.addWidget(Typography(text=children))
                self._layout.addWidget(QLabel(children))

        # === Style ===
        self._shadow_effect = None
        self._update_style()

    # ======================================================
    # =============== Internal Helpers =====================
    # ======================================================

    def _blur_for_elevation(self, elevation: int) -> int:
        """Map elevation (0–24) to shadow blur."""
        return max(2, min(64, elevation * 3))

    def _update_style(self, component_styled=None):
        self.theme = useTheme()
        ownerState = {}
        if not component_styled:
            component_styled = self.theme.components

        PyPaper_root = component_styled["PyPaper"].get("styles")["root"]
        PyPaper_root_qss = get_qss_style(PyPaper_root)
        
        PyPaper_outlined = component_styled["PyPaper"].get("styles")["outlined"]
        PyPaper_outlined_qss = get_qss_style(PyPaper_outlined)
        
        
        # Apply base style
        if self._variant == "outlined":
            self.setProperty("outlined", True)
            
        square_style = ""
        if self._square:
            square_style = "border-radius: 0px;"
            
        color_style = ""
        if self._color:
            color_style = f"background-color: {self._color};"
            
        
        sx_qss = ""
        if self._sx:
            if isinstance(self._sx, State):
                sx = self._sx.value
            elif isinstance(self._sx, Callable):
                sx = self._sx()
            else:
                sx = self._sx

            if isinstance(sx, dict):
                sx_qss = get_qss_style(sx, class_name=f"#{self.objectName()}")
            elif isinstance(sx, str) and sx != "":
                sx_qss = sx

        stylesheet = f"""
            #{self.objectName()} {{
                {PyPaper_root_qss}
                {square_style}
                {color_style}
            }}
            #{self.objectName()}[outlined=true] {{
                {PyPaper_outlined_qss}
            }}
            
            {sx_qss}
        """
        self.setStyleSheet(stylesheet)

        # Apply shadow only if variant == elevation and elevation > 0
        if self._variant == "elevation" and self._elevation > 0:
            if not self._shadow_effect:
                self._shadow_effect = QGraphicsDropShadowEffect(self)
            self._shadow_effect.setOffset(0, self._elevation)
            self._shadow_effect.setBlurRadius(self._blur_for_elevation(self._elevation))
            self._shadow_effect.setColor(rgbaToQColor(self._color) if self._color else QColor(0, 0, 0, 30 + self._elevation * 5))
            # self._shadow_effect.setColor(QColor(0, 0, 0, 10 + self._elevation * 3))
            self.setGraphicsEffect(self._shadow_effect)
        else:
            self.setGraphicsEffect(None)

    # ======================================================
    # =============== Getters / Setters ====================
    # ======================================================

    def getElevation(self):
        return self._elevation

    def setElevation(self, value):
        self._elevation = max(0, min(24, value))
        self._update_style()

    elevation = Property(int, getElevation, setElevation)

    def getVariant(self):
        return self._variant

    def setVariant(self, value):
        self._variant = value
        self._update_style()

    variant = Property(str, getVariant, setVariant)

    def getSquare(self):
        return self._square

    def setSquare(self, value: bool):
        self._square = value
        self._radius = 0 if value else 8
        self._update_style()

    square = Property(bool, getSquare, setSquare)

    def getColor(self):
        return self._color

    def setColor(self, color):
        self._color = QColor(color)
        self._update_style()

    color = Property(QColor, getColor, setColor)

    def getSX(self):
        return self._sx

    def setSX(self, sx):
        self._sx = sx
        self._update_style()

    sx = Property("QVariant", getSX, setSX)



import sys
import math
import re

from PySide6.QtWidgets import (
    QApplication,
    QWidget,
    QPushButton,
    QVBoxLayout,
    QGraphicsEffect,
    QGraphicsScene,
    QGraphicsPixmapItem,
    QGraphicsBlurEffect,
)
from PySide6.QtCore import (
    Qt,
    QRectF,
    QPointF,
    Property,
    QPropertyAnimation,
    QEasingCurve,
    QTimer,
    QObject,
)
from PySide6.QtGui import (
    QPainter,
    QColor,
    QPainterPath,
    QConicalGradient,
    QPixmap,
    QImage,
)

# ============================================================
# Parse CSS box-shadow (string) -> BoxShadowSpec
# Hỗ trợ:
#   "10px 10px grey"
#   "20px 20px 50px 15px grey"
#   "... pink inset"
#   "none"
# ============================================================

_COLOR_KEYWORDS = {
    "black": QColor(0, 0, 0),
    "white": QColor(255, 255, 255),
    "red": QColor(255, 0, 0),
    "green": QColor(0, 128, 0),
    "blue": QColor(0, 0, 255),
    "pink": QColor(255, 192, 203),
    "grey": QColor(128, 128, 128),
    "gray": QColor(128, 128, 128),
}


def _parse_px(token: str):
    m = re.fullmatch(r"(-?\d+)\s*px", token.strip().lower())
    return int(m.group(1)) if m else None


def _parse_color(token: str):
    t = token.strip().lower()

    if t in _COLOR_KEYWORDS:
        return QColor(_COLOR_KEYWORDS[t])

    # #RRGGBB / #AARRGGBB
    if re.fullmatch(r"#([0-9a-f]{6}|[0-9a-f]{8})", t):
        c = QColor(t)
        return c if c.isValid() else None

    # rgb(...)
    m = re.fullmatch(r"rgb\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)", t)
    if m:
        return QColor(int(m.group(1)), int(m.group(2)), int(m.group(3)))

    # rgba(..., a)  a: 0..1 hoặc 0..255
    m = re.fullmatch(
        r"rgba\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*,\s*([0-9]*\.?[0-9]+)\s*\)", t
    )
    if m:
        r, g, b = int(m.group(1)), int(m.group(2)), int(m.group(3))
        a_raw = float(m.group(4))
        a = int(a_raw * 255) if a_raw <= 1.0 else int(a_raw)
        return QColor(r, g, b, max(0, min(255, a)))

    return None


class BoxShadowSpec:
    """
    Tương đương CSS:
      box-shadow: x y blur spread color [inset];
    """
    def __init__(self, x=0, y=0, blur=0, spread=0, color=None, inset=False):
        self.x = int(x)
        self.y = int(y)
        self.blur = int(blur)
        self.spread = int(spread)
        self.color = QColor(color) if color is not None else QColor(0, 0, 0, 140)
        self.inset = bool(inset)


def parse_box_shadow(value: str):
    v = (value or "").strip().rstrip(";").strip()
    if not v or v.lower() == "none":
        return None

    tokens = [t for t in v.split() if t]

    inset = False
    if "inset" in [t.lower() for t in tokens]:
        inset = True
        tokens = [t for t in tokens if t.lower() != "inset"]

    # Lấy token màu đầu tiên parse được
    color = None
    rest = []
    for t in tokens:
        c = _parse_color(t)
        if c is not None and color is None:
            color = c
        else:
            rest.append(t)

    if color is None:
        color = QColor(0, 0, 0, 140)

    # Lấy px theo thứ tự: x y [blur] [spread]
    nums = []
    for t in rest:
        px = _parse_px(t)
        if px is not None:
            nums.append(px)

    if len(nums) < 2:
        return None

    x, y = nums[0], nums[1]
    blur = nums[2] if len(nums) >= 3 else 0
    spread = nums[3] if len(nums) >= 4 else 0

    # CSS keyword thường opaque -> set alpha nhìn giống web hơn
    if color.alpha() == 255:
        color.setAlpha(150)

    return BoxShadowSpec(x=x, y=y, blur=blur, spread=spread, color=color, inset=inset)


# ============================================================
# AnimEffect – compositor QtMUI:
# - OUTSET: bake pixmap cache (mask -> blur offscreen -> tô màu)
# - BLUR=0 vẫn phải ra shadow "cứng" (đúng CSS)
# - INSET: vẽ nhẹ trong draw()
# ============================================================
class ShadowEffect:
    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(*args, **kwargs)  # ✅ cho phép cooperative init (an toàn)
        
        self._parent = None

        # Gradient border
        self._shadowBorder = 2
        self._shadowRadius = 0



        # Shadow spec
        self._shadowSpec = None

        # Match CSS
        self._shadowBlurMap = 0.85
        self._shadowStrength = 1.25

        # Cache outset shadow
        self._shadowCachePixmap = None
        self._shadowCacheRect = QRectF()
        self._shadowCacheKey = None
        self._shadowDirty = True
        self._rebuildScheduled = False

    def _initShadow(self, parent):
      self._parent = parent

    # =========================
    # API set box-shadow string
    # =========================
    def setBoxShadow(self, value: str):
        self._shadowSpec = parse_box_shadow(value)
        self._shadowDirty = True
        self._shadowCachePixmap = None
        self._shadowCacheKey = None

        if not self._rebuildScheduled:
            self._rebuildScheduled = True
            QTimer.singleShot(0, self._rebuild_shadow_cache)

        self.update()

    # =====================================================
    # boundingRectFor:
    # - OUTSET shadow dù blur=0 vẫn phải expand theo offset/spread
    # - INSET không cần expand
    # =====================================================
    def boundingRectFor(self, rect: QRectF) -> QRectF:
        b = max(0, self._shadowBorder)
        s = self._shadowSpec

        if s is None:
            return rect.adjusted(-b, -b, b, b)

        if s.inset or s.color.alpha() <= 0:
            return rect.adjusted(-b, -b, b, b)

        # OUTSET: blur có thể = 0 nhưng vẫn phải mở theo offset/spread
        expand = b + max(0, s.spread) + max(0, s.blur) + max(abs(s.x), abs(s.y))
        return rect.adjusted(-expand, -expand, expand, expand)



    # =====================================================
    # DPR (HiDPI)
    # =====================================================
    def _get_dpr(self) -> float:
        w = self._parent
        if w is not None and hasattr(w, "devicePixelRatioF"):
            try:
                return float(w.devicePixelRatioF())
            except Exception:
                return 1.0
        return 1.0

    # =====================================================
    # Cache key (outset)
    # =====================================================
    def _make_shadow_key(self, src_rect: QRectF, dpr: float, s: BoxShadowSpec) -> tuple:
        return (
            round(src_rect.width(), 2),
            round(src_rect.height(), 2),
            self._shadowRadius,
            s.x,
            s.y,
            s.blur,
            s.spread,
            s.color.rgba(),
            round(dpr, 2),
            round(self._shadowBlurMap, 3),
            round(self._shadowStrength, 3),
        )

    # =====================================================
    # Rebuild OUTSET shadow cache (ngoài draw)
    # - BLUR=0: không blur mask, vẫn tô màu và offset đúng CSS
    # =====================================================
    def _rebuild_shadow_cache(self):
      
      self._rebuildScheduled = False

      src_rect = self.sourceBoundingRect()
      if src_rect.isNull() or src_rect.width() <= 0 or src_rect.height() <= 0:
          return

      s = self._shadowSpec
      dpr = self._get_dpr()

      # Không có shadow hoặc inset -> không build cache
      if s is None or s.inset or s.color.alpha() <= 0:
          self._shadowCachePixmap = None
          self._shadowCacheKey = None
          self._shadowDirty = False
          self.update()
          return

      key = self._make_shadow_key(src_rect, dpr, s)
      if (not self._shadowDirty) and (key == self._shadowCacheKey):
          return

      self._shadowCacheKey = key
      self._shadowDirty = False

      full_rect = self.boundingRectFor(src_rect)
      self._shadowCacheRect = full_rect

      img_w = max(1, int(math.ceil(full_rect.width() * dpr)))
      img_h = max(1, int(math.ceil(full_rect.height() * dpr)))

      # 1) mask trắng (shape + spread + offset)
      mask = QImage(img_w, img_h, QImage.Format_ARGB32_Premultiplied)
      mask.fill(Qt.transparent)

      pmask = QPainter(mask)
      try:
          pmask.setRenderHint(QPainter.Antialiasing, True)
          pmask.scale(dpr, dpr)

          origin = QPointF(-full_rect.left(), -full_rect.top())

          base_rect = src_rect.adjusted(
              -s.spread, -s.spread, s.spread, s.spread
          ).translated(s.x, s.y).translated(origin)

          path = QPainterPath()
          path.addRoundedRect(
              base_rect,
              self._shadowRadius + s.spread,
              self._shadowRadius + s.spread,
          )
          pmask.fillPath(path, QColor(255, 255, 255, 255))
      finally:
          pmask.end()

      # 2) Nếu blur > 0 thì blur mask offscreen, còn blur == 0 thì dùng mask luôn
      if s.blur > 0:
          mask_pm = QPixmap.fromImage(mask)
          mask_pm.setDevicePixelRatio(dpr)

          scene = QGraphicsScene()
          item = QGraphicsPixmapItem(mask_pm)

          blur = QGraphicsBlurEffect()
          blur.setBlurRadius(float(s.blur) * float(self._shadowBlurMap))
          blur.setBlurHints(QGraphicsBlurEffect.QualityHint)

          item.setGraphicsEffect(blur)
          scene.addItem(item)

          blurred = QImage(img_w, img_h, QImage.Format_ARGB32_Premultiplied)
          blurred.fill(Qt.transparent)

          pblur = QPainter(blurred)
          try:
              scene.setSceneRect(0, 0, full_rect.width(), full_rect.height())
              scene.render(pblur, QRectF(0, 0, full_rect.width(), full_rect.height()))
          finally:
              pblur.end()
      else:
          # BLUR=0: mask chính là "blurred" (bóng cứng, đúng CSS)
          blurred = mask

      # 3) tô màu + strength
      boosted = QColor(s.color)
      boosted_alpha = int(boosted.alpha() * float(self._shadowStrength))
      boosted.setAlpha(max(0, min(255, boosted_alpha)))

      colored = QImage(img_w, img_h, QImage.Format_ARGB32_Premultiplied)
      colored.fill(Qt.transparent)

      pcol = QPainter(colored)
      try:
          pcol.fillRect(0, 0, img_w, img_h, boosted)
          pcol.setCompositionMode(QPainter.CompositionMode_DestinationIn)
          pcol.drawImage(0, 0, blurred)
      finally:
          pcol.end()

      pm = QPixmap.fromImage(colored)
      pm.setDevicePixelRatio(dpr)
      self._shadowCachePixmap = pm
      self.update()

    # =====================================================
    # Vẽ INSET shadow (nhẹ)
    # =====================================================
    def _draw_inset_shadow(self, painter: QPainter, src_rect: QRectF, s: BoxShadowSpec):
        if s.color.alpha() <= 0:
            return

        clip_path = QPainterPath()
        clip_path.addRoundedRect(src_rect, self._shadowRadius, self._shadowRadius)

        painter.save()
        try:
            painter.setClipPath(clip_path)

            # blur=0 inset: vẽ một lớp cứng ở trong
            if s.blur <= 0:
                rr = src_rect.adjusted(s.spread, s.spread, -s.spread, -s.spread).translated(s.x, s.y)
                path = QPainterPath()
                path.addRoundedRect(rr, max(0, self._shadowRadius - s.spread), max(0, self._shadowRadius - s.spread))
                painter.fillPath(path, s.color)
                return

            # blur>0 inset: vài vòng nhẹ
            steps = min(48, max(18, int(s.blur * 0.9)))
            base_rect = src_rect.adjusted(s.spread, s.spread, -s.spread, -s.spread).translated(s.x, s.y)

            for i in range(steps):
                t = i / steps
                alpha = int(s.color.alpha() * (1.0 - t) ** 2)
                if alpha <= 0:
                    continue

                c = QColor(s.color)
                c.setAlpha(alpha)

                inset = t * s.blur
                rr = base_rect.adjusted(inset, inset, -inset, -inset)
                if rr.width() <= 0 or rr.height() <= 0:
                    break

                path = QPainterPath()
                path.addRoundedRect(rr, max(0.0, self._shadowRadius - inset), max(0.0, self._shadowRadius - inset))
                painter.fillPath(path, c)
        finally:
            painter.restore()




from PySide6.QtWidgets import QWidget, QApplication
from PySide6.QtCore import QTimer
from PySide6.QtGui import QPainter, QColor
import sys
import random


from qtmui.material.styles import useTheme

COLORS = ['default', 'primary', 'secondary', 'info', 'success', 'warning', 'error']

def alpha(color: QColor, alpha: float) -> QColor:
    """Trả về QColor với alpha (0..1)."""
    c = QColor(color)
    c.setAlphaF(alpha)  # hoặc setAlpha(int(alpha*255))
    return c

class LinearDeterminate(QWidget):
    def __init__(
        self,
        key: str = None,
        color: str = "default",
        value: float = 0.0,
    ):
        super().__init__()
        self.setMinimumHeight(5)
        self.setMinimumWidth(300)

        # Progress giống MUI
        self.progress = value

        # Màu sắc
        self.theme = useTheme()
        
        if color == "default":
            self.value_color = self.theme.palette.text.secondary
        elif color in COLORS:
            self.value_color = getattr(self.theme.palette, color).main
        
        self.bg_color = alpha(self.value_color, 0.16)


        # Timer mô phỏng MUI progress
        self.progress_timer = QTimer()
        self.progress_timer.timeout.connect(self.update_progress)
        self.progress_timer.start(500)

        # Timer refresh vẽ
        self.render_timer = QTimer()
        self.render_timer.timeout.connect(self.update)
        self.render_timer.start(16)  # ~60 FPS

    def update_progress(self):
        # Random increment giống MUI
        diff = random.uniform(0, 10)
        if self.progress + diff > 100:
            self.progress = 0
        else:
            self.progress += diff

    def paintEvent(self, event):
        w = self.width()
        h = self.height()
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        # Nền
        painter.fillRect(0, 0, w, h, self.bg_color)

        # Value
        value_width = w * (self.progress / 100)
        painter.fillRect(0, 0, value_width, h, self.value_color)

# if __name__ == "__main__":
#     app = QApplication(sys.argv)
#     w = LinearDeterminate()
#     w.setWindowTitle("Linear Determinate Progress")
#     w.show()
#     sys.exit(app.exec())



from typing import Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from qtmui.material.styles.create_theme.theme_reducer import ThemeState

from .properties_name import *
from ....system.color_manipulator import alpha

class ChipClasses:
    deleteIcon: str = "MuiChip-deleteIcon"
    avatar: str = "MuiChip-avatar"
    icon: str = "MuiChip-icon"
    disabled: str = "Mui-disabled"


class ChipProps:
    color: str
    variant: str


class ChipStyle:
    muiChip: Dict = {}


def chip(_theme):
    theme: ThemeState = _theme

    lightMode = theme.palette.mode == 'light'
    COLORS = ['primary', 'secondary', 'info', 'success', 'warning', 'error']

    return {
        'PyChip': {
            'styles': {
                'root': {
                    **{
                        "default": {
                            color: "red",
                            "fontWeight": 500,
                            "borderRadius": f"{theme.shape.borderRadius}px",
                            "textContent": {
                                fontSize: theme.typography.button.fontSize,
                                fontWeight: theme.typography.button.fontWeight,
                                lineHeight: theme.typography.button.lineHeight,
                            },
                            "deleteIcon": {
                                'size': 16,
                                'border-radius': '8px',
                                "props": {
                                    "filledVariant": {
                                        backgroundColor: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        "hover": {
                                            backgroundColor: alpha(theme.palette.common.white, 0.24) if lightMode else alpha(theme.palette.grey._800, 0.24),
                                        }
                                    },
                                    "outlinedVariant": {
                                        backgroundColor: theme.palette.grey._500,
                                        "hover": {
                                            backgroundColor: theme.palette.common.black if lightMode else theme.palette.common.white,
                                        }
                                    },
                                    "softVariant": {
                                        backgroundColor: theme.palette.grey._500,
                                        "hover": {
                                            backgroundColor: theme.palette.common.black if lightMode else theme.palette.common.white,
                                        }
                                    },
                                },
                            },
                            "slots": {
                                "hover": {
                                    backgroundColor: "transparent"
                                }
                            },
                            "props": {
                                "smallSize": {
                                    p: "1px 2px"
                                },
                                "mediumSize": {
                                    p: "2px 4px"
                                },
                                "filledVariant": {
                                    color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                    backgroundColor: theme.palette.text.primary,
                                    "slots": {
                                        "hover": {
                                            backgroundColor: theme.palette.grey._700 if lightMode else theme.palette.grey._100,
                                        }
                                    },
                                    "classes": {
                                        "icon": {
                                            color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        }
                                    }
                                },
                                "outlinedVariant": {
                                    color: theme.palette.grey._800 if lightMode else theme.palette.common.white,
                                    backgroundColor: "transparent",
                                    border: f"1px solid {alpha(theme.palette.grey._500, 0.16)}",
                                    "slots": {
                                        "hover": {
                                            backgroundColor: "transparent"
                                        }
                                    },
                                    "classes": {
                                        "icon": {
                                            color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        }
                                    }
                                },
                                "softVariant": {
                                    color: theme.palette.text.primary,
                                    backgroundColor: alpha(theme.palette.grey._500, 0.16),
                                    "slots": {
                                        "hover": {
                                            backgroundColor: alpha(theme.palette.grey._500, 0.32),
                                        }
                                    },
                                    "classes": {
                                        "icon": {
                                            color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        }
                                    }
                                },

                            },
                        }
                    },
                    **{
                        f"{_color}": {
                            color: "red",
                            "fontWeight": 500,
                            "border": "1px solid transparent",
                            "borderRadius": f"{theme.shape.borderRadius}px",
                            "textContent": {
                                fontSize: theme.typography.button.fontSize,
                                fontWeight: theme.typography.button.fontWeight,
                                lineHeight: theme.typography.button.lineHeight,
                            },
                            "deleteIcon": {
                                borderRadius: '8px',
                                'size': 16,
                                "props": {
                                    "filledVariant": {
                                        backgroundColor: alpha(theme.palette.common.white, 0.64) if lightMode else alpha(theme.palette.grey._800, 0.64),
                                        "hover": {
                                            backgroundColor: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        }
                                    },
                                    "outlinedVariant": {
                                        backgroundColor: alpha(getattr(theme.palette, _color).light, 0.64),
                                        "hover": {
                                            backgroundColor: getattr(theme.palette, _color).light,
                                        }
                                    },
                                    "softVariant": {
                                        backgroundColor: alpha(getattr(theme.palette, _color).dark, 0.64),
                                        "hover": {
                                            backgroundColor: getattr(theme.palette, _color).dark,
                                        }
                                    },
                                },
                            },
                            "slots": {
                                "hover": {
                                }
                            },
                            "props": {
                                "smallSize": {
                                    p: "1px 2px"
                                },
                                "mediumSize": {
                                    p: "2px 4px"
                                },
                                "filledVariant": {
                                    color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                    backgroundColor: getattr(theme.palette, _color).main,
                                    "slots": {
                                        "hover": {
                                            backgroundColor: getattr(theme.palette, _color).dark,
                                        }
                                    },
                                    "classes": {
                                        "icon": {
                                            color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        }
                                    }
                                },
                                "outlinedVariant": {
                                    color: getattr(theme.palette, _color).dark if lightMode else getattr(theme.palette, _color).light,
                                    backgroundColor: "transparent",
                                    borderColor: getattr(theme.palette, _color).main,
                                    "slots": {
                                        "hover": {
                                            backgroundColor: alpha(getattr(theme.palette, _color).main, 0.08),
                                        }
                                    },
                                    "classes": {
                                        "icon": {
                                            color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        }
                                    }
                                },
                                "softVariant": {
                                    color: getattr(theme.palette, _color).dark if lightMode else getattr(theme.palette, _color).light,
                                    backgroundColor: alpha(getattr(theme.palette, _color).main, 0.16),
                                    "slots": {
                                        "hover": {
                                            backgroundColor: alpha(getattr(theme.palette, _color).main, 0.32),
                                        }
                                    },
                                    "classes": {
                                        "icon": {
                                            color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                                        }
                                    }
                                },

                            },
                        } 
                        for _color in COLORS
                    }
                },
                "avatar": {
                    color: theme.palette.text.primary,
                },
                "icon": {

                }
            },
        },

    }



from typing import Callable
from functools import wraps
import json

# Hàm để merge dict tương tự lodash.merge
def merge_dicts(*dict_args)-> dict:
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

def deep_merge(a, b):
    """
    Gộp hai giá trị a và b.
    - Nếu cả a và b đều là dict, thực hiện gộp sâu các key.
    - Nếu cả a và b đều callable, tạo một hàm mới gọi cả hai hàm và merge kết quả trả về.
    - Ngược lại, ưu tiên giá trị b (giá trị ghi đè).
    """
    # Nếu cả hai đều là dict, gộp chúng theo cách đệ quy:
    if isinstance(a, dict) and isinstance(b, dict):
        result = a.copy()
        for key, value in b.items():
            if key in result:
                result[key] = deep_merge(result[key], value)
            else:
                result[key] = value
        return result
    # Nếu cả hai đều callable:
    if callable(a) and callable(b):
        return lambda *args, **kwargs: deep_merge(a(*args, **kwargs), b(*args, **kwargs))
    # Trường hợp còn lại: ưu tiên b (gọi là override)
    return b if b is not None else a


def convert_sx_params_to_str(func):
    @wraps(func)
    def wrapper(sx, *args, **kwargs):
        # Nếu params là dict, chuyển nó thành chuỗi JSON (hoặc sử dụng str() nếu phù hợp)
        if isinstance(sx, dict):
            sx = json.dumps(sx)
        elif isinstance(sx, Callable):
            sx = sx()
            sx = json.dumps(sx)
        else:
            if sx == None:
                sx = ""
        return func(sx, *args, **kwargs)
    return wrapper

def convert_sx_params_to_dict(func):
    @wraps(func)
    def wrapper(sx, *args, **kwargs):
        # Nếu params là dict, chuyển nó thành chuỗi JSON (hoặc sử dụng str() nếu phù hợp)
        if isinstance(sx, str) and sx != "":
            try:
                sx = json.loads(sx)
            except Exception as e:
                print('sxxxxxxxxxx_errrrrrrrrrrrrr', type(sx))
        return func(sx, *args, **kwargs)
    return wrapper





# src/qtmui/material/image_list_item.py
from typing import Optional, Union, Callable, Dict, List
from functools import lru_cache
from typing import TYPE_CHECKING
from PySide6.QtWidgets import QFrame, QVBoxLayout
from PySide6.QtCore import Qt
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style
from qtmui.hooks import State

if TYPE_CHECKING:
    from ..image import Image
    from ..image_list_item_bar import ImageListItemBar

class ImageListItem(QFrame):
    def __init__(
        self,
        children=None,
        cols=1,
        rows=1,
        title=None,
        subtitle=None,
        actionIcon=None,
        titlePosition="bottom",  # top, bottom, below
        sx: Optional[Union[State, Callable, str, Dict]] = None,
        **kwargs
    ):
        super().__init__()
        self.setObjectName(f"ImageListItem{str(id(self))}")
        if sx:
            self._setSx(sx)
            
        self._kwargs = {
            "cols": cols,
            "rows": rows,
            "title": title,
            "subtitle": subtitle,
            "actionIcon": actionIcon,
            "titlePosition": titlePosition,
            **kwargs
        }
        self._setKwargs(self._kwargs)
        
        self.cols = cols
        self.rows = rows
        self._sx = sx

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        # Container cho image + overlay
        self.image_container = QFrame()
        # self.image_container.setStyleSheet("background: #000; border-radius: 8px; overflow: hidden;")
        self.image_container.setStyleSheet("background: #000; border-radius: 8px")
        container_layout = QVBoxLayout(self.image_container)
        container_layout.setContentsMargins(0, 0, 0, 0)
        container_layout.setSpacing(0)

        if children:
            for child in children:
                # if isinstance(child, Image):
                #     child.setStyleSheet("border-radius: 8px;")
                container_layout.addWidget(child)

        layout.addWidget(self.image_container)

        # Thêm titlebar nếu có
        if title or subtitle or actionIcon:
            bar = ImageListItemBar(
                title=title or "",
                subtitle=subtitle,
                actionIcon=actionIcon,
                position=titlePosition,
            )
            if titlePosition in ["top", "bottom"]:
                # Overlay vào trong image
                container_layout.addWidget(bar)
                bar.setStyleSheet(bar.styleSheet() + "position: absolute; width: 100%;")
                bar.setAttribute(Qt.WA_TransparentForMouseEvents, titlePosition == "bottom")
            else:
                # Below image
                layout.addWidget(bar)

        self.setLayout(layout)
        
        self._setStyleSheet()
        
    @classmethod
    def _setSx(cls, sx: dict = {}):
        cls.sxDict = sx
        
    @classmethod
    def _setKwargs(cls, kwargs: dict = {}):
        cls.ownerState = kwargs

    @classmethod
    @lru_cache(maxsize=128)
    def _getSxQss(cls, sxStr: str = "", className: str = "PyWidgetBase"):
        sx_qss = get_qss_style(cls.sxDict, class_name=className)
        return sx_qss
        
    def _setStyleSheet(self):
        
        # stylesheet = ""
        # if hasattr(self, "styledDict"):
        #     root = self.styledDict.get("PyBox", {}).get("styles", {}).get("root", None)(self._kwargs)
        #     if root:
        #         stylesheet = self._getStyleSheet(styledConfig=str(root))
        # else:
        #     stylesheet = self._getStyleSheet()
            
        sxQss = ""
        if self._sx:
            sxQss = self._getSxQss(sxStr=str(self._sx), className=f"#{self.objectName()}")

        stylesheet = f"""
            {sxQss}
        """
        
        self.setStyleSheet(stylesheet)


from typing import Literal, Tuple, Union, Dict
import re
from PySide6.QtGui import QColor
import random


# Định nghĩa kiểu ColorFormat
ColorFormat = Literal['rgb', 'rgba', 'hsl', 'hsla', 'color']

# Định nghĩa kiểu ColorSpace
ColorSpace = Literal['srgb', 'display-p3', 'a98-rgb', 'prophoto-rgb', 'rec-2020']

# Định nghĩa kiểu ColorObject
class ColorObject:
    def __init__(
        self,
        type: ColorFormat,
        values: Union[Tuple[int, int, int], Tuple[int, int, int, int]],
        color_space: ColorSpace = 'srgb'
    ):
        self.type = type
        self.values = values
        self.color_space = color_space

    def __repr__(self):
        return f"ColorObject(type={self.type}, values={self.values}, color_space={self.color_space})"


def clamp(value, min_value=0, max_value=1):
    if value < min_value or value > max_value:
        print(f"Warning: The value provided {value} is out of range [{min_value}, {max_value}].")
    return max(min_value, min(max_value, value))

def hexToRgb(color):
    color = color.lstrip('#')
    length = 2 if len(color) >= 6 else 1
    colors = re.findall('.' * length, color)
    if len(colors) == 3:
        return f"rgb({', '.join(str(int(c, 16)) for c in colors)})".encode('utf-8').decode('utf-8')
    elif len(colors) == 4:
        return f"rgba({', '.join(str(int(c, 16)) for c in colors[:3])}, {round(int(colors[3], 16) / 255, 3)})".encode('utf-8').decode('utf-8')
    return ''

def int_to_hex(value):
    hex_value = hex(value)[2:]
    return hex_value if len(hex_value) > 1 else f'0{hex_value}'


# Từ điển chuẩn hóa tên màu CSS → mã hex
CSS_COLOR_NAMES = {
    "black": "#000000",
    "white": "#ffffff",
    "red": "#ff0000",
    "lime": "#00ff00",
    "blue": "#0000ff",
    "yellow": "#ffff00",
    "cyan": "#00ffff",
    "magenta": "#ff00ff",
    "silver": "#c0c0c0",
    "gray": "#808080",
    "maroon": "#800000",
    "olive": "#808000",
    "green": "#008000",
    "purple": "#800080",
    "teal": "#008080",
    "navy": "#000080",
    "orange": "#ffa500",
    "pink": "#ffc0cb",
    "brown": "#a52a2a",
    "gold": "#ffd700",
    "indigo": "#4b0082",
    "violet": "#ee82ee",
    "lightgray": "#d3d3d3",
    "darkgray": "#a9a9a9",
    # bạn có thể thêm các màu khác nếu cần
}


def hexToRgb(hex_color: str):
    """Chuyển mã hex → rgb tuple."""
    hex_color = hex_color.lstrip("#")
    if len(hex_color) == 3:
        hex_color = "".join([c * 2 for c in hex_color])
    r, g, b = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))
    return f"rgb({r}, {g}, {b})"


def decomposeColor(color: Union[str, Dict]) -> Dict:
    """Phân tích mọi dạng màu: tên, hex, rgb, rgba, hsl, hsla, hoặc color()."""
    color = str(color).strip().lower()

    # --- 1️⃣ Nếu là tên màu CSS chuẩn (vd: "red", "blue") ---
    if color in CSS_COLOR_NAMES:
        return decomposeColor(CSS_COLOR_NAMES[color])

    # --- 2️⃣ Nếu là mã hex ---
    if color.startswith("#"):
        return decomposeColor(hexToRgb(color))

    # --- 3️⃣ Nếu là dạng rgb(), rgba(), hsl(), hsla(), color() ---
    marker = color.find("(")
    if marker == -1:
        raise ValueError(f"Unsupported color format: {color}")

    color_type = color[:marker].strip()
    values = color[marker + 1 : -1].split(",")

    if color_type not in ["rgb", "rgba", "hsl", "hsla", "color"]:
        raise ValueError(f"Unsupported color type: {color_type}")

    # Xử lý từng giá trị (loại bỏ % nếu có)
    parsed_values = []
    for v in values:
        v = v.strip()
        if v.endswith("%"):
            parsed_values.append(float(v[:-1]))  # bỏ ký hiệu %
        else:
            try:
                parsed_values.append(float(v))
            except ValueError:
                parsed_values.append(v)  # nếu là string (vd 'none')

    return {
        "type": color_type,
        "values": parsed_values
    }


def recomposeColor(color):
    color_type = color['type']
    values = color['values']
    
    if color_type.startswith('rgb'):
        values = [int(v) for v in values[:3]] + values[3:]
    elif color_type.startswith('hsl'):
        values[1] = f"{values[1]}%"
        values[2] = f"{values[2]}%"
    
    return f"{color_type}({', '.join(map(str, values))})".encode('utf-8').decode('utf-8')

def hslToRgb(color):
    """
    Chuyển đổi màu HSL sang RGB.
    """
    color = decomposeColor(color)
    values = color["values"]
    h = values[0]
    s = values[1] / 100
    l = values[2] / 100

    a = s * min(l, 1 - l)

    def f(n):
        k = (n + h / 30) % 12
        return l - a * max(min(k - 3, 9 - k, 1), -1)

    rgb = [
        round(f(0) * 255),
        round(f(8) * 255),
        round(f(4) * 255),
    ]

    if color["type"] == "hsla":
        rgb.append(values[3])  # Thêm alpha nếu là HSLA

    return recomposeColor({
        "type": "rgba" if color["type"] == "hsla" else "rgb",
        "values": rgb,
    })



def rgbToHex(color):
    decomposed = decomposeColor(color)
    return '#' + ''.join(int_to_hex(int(v)) for v in decomposed['values'][:3])

def rgb2hex(r_or_color: Union[tuple, int], g: int = 0, b: int = 0, a: int = 0) -> str:
    """Convert rgb color to hex color.

    :param r_or_color: The 'red' value or a color tuple.
    :param g: The 'green' value.
    :param b: The 'blue' value.
    :param a: The 'alpha' value.
    :return: The converted hexadecimal color.
    """

    if type(r_or_color).__name__ == "tuple": r, g, b = r_or_color[:3]
    else: r = r_or_color
    hex = '%02x%02x%02x' % (int(r), int(g), int(b))
    return hex

def rgbaToQColor(rgba_string):
    rgba_values = rgba_string.strip("rgba()").split(", ")
    rgba_values = [float(value.strip()) for value in rgba_values]

    if len(rgba_values) == 4:
        red = rgba_values[0] / 255.0
        green = rgba_values[1] / 255.0
        blue = rgba_values[2] / 255.0
        alpha = rgba_values[3]

        return QColor.fromRgbF(red, green, blue, alpha)
    return None

def getLuminance(color):
    decomposed = decomposeColor(color)
    rgb = decomposed['values'][:3]
    rgb = [v / 255 for v in rgb]
    rgb = [(v / 12.92 if v <= 0.03928 else ((v + 0.055) / 1.055) ** 2.4) for v in rgb]
    return round(0.2126 * rgb[0] + 0.7152 * rgb[1] + 0.0722 * rgb[2], 3)

def getContrastRatio(foreground, background):
    lum_a = getLuminance(foreground)
    lum_b = getLuminance(background)
    return (max(lum_a, lum_b) + 0.05) / (min(lum_a, lum_b) + 0.05)

def alpha_tuple(color, value):
    decomposed = decomposeColor(color)
    value = clamp(value)
    decomposed['values'].append(value)
    return recomposeColor(decomposed)

def alpha(color, value):
    decomposed_color = decomposeColor(color)
    value = clamp(value)

    # Kiểm tra nếu chưa có giá trị alpha, thêm nó vào
    if len(decomposed_color['values']) == 3:
        decomposed_color['values'].append(value)  # Thêm kênh alpha mới
    else:
        decomposed_color['values'][3] = value  # Cập nhật kênh alpha hiện có
    
    if decomposed_color['type'] == 'rgb' or decomposed_color['type'] == 'hsl':
        decomposed_color['type'] += 'a'  # Chuyển đổi sang rgba hoặc hsla
    
    return recomposeColor(decomposed_color)


def darken(color, coefficient):
    decomposed = decomposeColor(color)
    coefficient = clamp(coefficient)
    
    for i in range(3):
        decomposed['values'][i] *= (1 - coefficient)
    
    return recomposeColor(decomposed)

def lighten(color, coefficient):
    decomposed = decomposeColor(color)
    coefficient = clamp(coefficient)
    
    for i in range(3):
        decomposed['values'][i] += (255 - decomposed['values'][i]) * coefficient
    
    return recomposeColor(decomposed)

def emphasize(color, coefficient=0.15):
    return darken(color, coefficient) if getLuminance(color) > 0.5 else lighten(color, coefficient)

def blend(background, overlay, opacity, gamma=1.0):
    def blend_channel(bg, ov):
        return round((bg ** (1 / gamma) * (1 - opacity) + ov ** (1 / gamma) * opacity) ** gamma)
    
    background_color = decomposeColor(background)
    overlay_color = decomposeColor(overlay)
    
    blended = [blend_channel(bg, ov) for bg, ov in zip(background_color['values'][:3], overlay_color['values'][:3])]
    
    return recomposeColor({'type': 'rgb', 'values': blended})

def hex_string_to_qcolor(hex_string: str):
    # Loại bỏ dấu '#' nếu có
    if hex_string.startswith('#'):
        hex_string = hex_string[1:]

    # Đảm bảo độ dài hợp lệ
    if len(hex_string) not in (6, 8):
        raise ValueError("Invalid hex color string length")

    # Tách giá trị RGB và alpha nếu có
    r = int(hex_string[0:2], 16)
    g = int(hex_string[2:4], 16)
    b = int(hex_string[4:6], 16)
    a = int(hex_string[6:8], 16) if len(hex_string) == 8 else 50

    return QColor(r, g, b, a)

def get_palette_text_color(palette, color):
    if color == 'primary':
        return palette.text.primary
    elif color == 'secondary':
        return palette.text.secondary
    # Thêm các màu khác ở đây
    return '#000000'  # Màu mặc định nếu không tìm thấy


def lighten_hex(hex_color, value):
    def _clamp(value, min_value=0, max_value=255):
        """Giữ giá trị trong khoảng từ min_value đến max_value."""
        return max(min_value, min(max_value, value))

    def _hexToRgb(_hex_color):
        """Chuyển mã màu hex thành RGB."""
        _hex_color = _hex_color.lstrip('#')
        return tuple(int(_hex_color[i:i+2], 16) for i in (0, 2, 4))

    def _rgbToHex(rgb):
        """Chuyển từ RGB thành mã màu hex."""
        return '#{:02x}{:02x}{:02x}'.format(_clamp(rgb[0]), _clamp(rgb[1]), _clamp(rgb[2]))
    """Làm nhạt màu hex và trả về kết quả dưới dạng mã hex."""
    # Chuyển hex sang RGB
    rgb = _hexToRgb(hex_color)

    # Làm sáng màu bằng cách tăng giá trị RGB
    lightened_rgb = tuple(_clamp(int(c + (255 - c) * value)) for c in rgb)

    # Chuyển ngược lại sang hex
    return _rgbToHex(lightened_rgb)


def interpolate_color(start_color: str, end_color: str, factor: float) -> str:
    """Nội suy màu từ start_color đến end_color dựa trên giá trị factor từ 0 đến 1."""
    start_color = start_color.lstrip('#')
    end_color = end_color.lstrip('#')

    # Tách thành giá trị RGB
    start_rgb = [int(start_color[i:i+2], 16) for i in (0, 2, 4)]
    end_rgb = [int(end_color[i:i+2], 16) for i in (0, 2, 4)]

    # Tính giá trị màu nội suy
    interpolated_rgb = [
        int(start_rgb[i] + (end_rgb[i] - start_rgb[i]) * factor) for i in range(3)
    ]

    return '#{:02X}{:02X}{:02X}'.format(*interpolated_rgb)

def generate_color_palette(start_color: str, end_color: str, steps: int) -> list:
    """Tạo dải màu từ start_color đến end_color với số lượng bước colors."""
    return [interpolate_color(start_color, end_color, i / (steps - 1)) for i in range(steps)]


def get_contrast_text(background, contrast_threshold=3):
    """Trả về màu văn bản có độ tương phản cao nhất với nền."""
    dark_text = "#ffffff"
    light_text = "#000000"

    contrast_text = dark_text if getContrastRatio(background, dark_text) >= contrast_threshold else light_text
    
    # Kiểm tra độ tương phản nếu môi trường không phải production
    contrast = getContrastRatio(background, contrast_text)
    if contrast < 3:
        print(f"The contrast ratio of {contrast}:1 for {contrast_text} on {background} falls below the WCAG recommended minimum contrast ratio of 3:1.")
    
    return contrast_text


def rgba_to_hex(r=None, g=None, b=None, a=None, rgba=None):
    if rgba:
        rgba = str(rgba).replace("rgba(", "").replace(")", "").strip(" ").split(',')
        r, g, b, a = int(rgba[0]), int(rgba[1]), int(rgba[2]), int(rgba[3])
    """
    Chuyển đổi giá trị RGBA thành chuỗi mã màu hex.
    Tham số:
        r (int): Giá trị đỏ (0-255)
        g (int): Giá trị xanh lá cây (0-255)
        b (int): Giá trị xanh dương (0-255)
        a (int): Giá trị alpha (0-255)
    Trả về:
        str: Chuỗi mã màu hex tương ứng.
    """
    # Kiểm tra giá trị đầu vào
    for val in (r, g, b, a):
        if not (isinstance(val, int) and 0 <= val <= 255):
            raise ValueError("Các giá trị RGBA phải là số nguyên từ 0 đến 255.")
    return '#{:02x}{:02x}{:02x}{:02x}'.format(r, g, b, a)


def get_flat_ui_color_hex(color_name):
    flat_ui_colors = {
        "turquoise": "#1abc9c",
        "green_sea": "#16a085",
        "emerald": "#2ecc71",
        "nephritis": "#27ae60",
        "peter_river": "#3498db",
        "belize_hole": "#2980b9",
        "amethyst": "#9b59b6",
        "wisteria": "#8e44ad",
        "wet_asphalt": "#34495e",
        "midnight_blue": "#2c3e50",
        "sunflower": "#f1c40f",
        "orange": "#f39c12",
        "carrot": "#e67e22",
        "pumpkin": "#d35400",
        "alizarin": "#e74c3c",
        "pomegranate": "#c0392b",
        "clouds": "#ecf0f1",
        "silver": "#bdc3c7",
        "concrete": "#95a5a6",
        "asbestos": "#7f8c8d"
    }
    
    # Trả về mã màu hex hoặc None nếu không tìm thấy màu
    return flat_ui_colors.get(color_name.lower(), None)

def get_random_flat_ui_color():
    flat_ui_colors = [
        "#1abc9c",  # Turquoise
        "#16a085",  # Green Sea
        "#2ecc71",  # Emerald
        "#27ae60",  # Nephritis
        "#3498db",  # Peter River
        "#2980b9",  # Belize Hole
        "#9b59b6",  # Amethyst
        "#8e44ad",  # Wisteria
        "#34495e",  # Wet Asphalt
        "#2c3e50",  # Midnight Blue
        "#f1c40f",  # Sunflower
        "#f39c12",  # Orange
        "#e67e22",  # Carrot
        "#d35400",  # Pumpkin
        "#e74c3c",  # Alizarin
        "#c0392b",  # Pomegranate
        "#ecf0f1",  # Clouds
        "#bdc3c7",  # Silver
        "#95a5a6",  # Concrete
        "#7f8c8d",  # Asbestos
    ]
    
    return random.choice(flat_ui_colors)

# # Ví dụ sử dụng:
# color_hex = get_flat_ui_color_hex("turquoise")
# print(color_hex)  # Output: #1abc9c

# # Màu đầu và màu cuối
# start_color = '#161C24'  # Màu đầu
# end_color = '#212B36'    # Màu cuối

# # Tạo dải màu gồm 4 màu (main, secondary, thirty, content)
# palette = generate_color_palette(start_color, end_color, 4)

# # In kết quả
# for i, color in enumerate(palette, 1):
#     print(f"Màu {i}: {color}")

# Example usage
# color1 = '#ff5733'
# color2 = '#333333'
# print(rgbToHex('rgb(255, 87, 51)'))
# print(darken(color1, 0.2))

# print(get_contrast_text('#004B50'))
